import pandas as pd
import numpy as np
from functools import reduce
import warnings


def check_fairness_groups(df_1, df_2, multiindex=True):
    """Check if two DataFrames have different group of fairness or if Multiindex Dataframes have different columns until level 1 and retrieves the list of unmatched fairness groups.

    Args:
        df_1 (pd.DataFrame): Multiindex DataFrame 1
        df_2 (pd.DataFrame): Multiindex DataFrame 2
        multiindex (bool, optional): determines if args are Multiindex Dataframes or DataFrames. Default to True.

    Returns:
        list: list of unmatched fairness groups unmatched.
    """
    if multiindex:
        col_2_train = reduce(
            lambda l, y: l.append(y) or l if y not in l else l, [(x[0], x[1]) for x in df_1.columns], []
        )
        col_2_test = reduce(
            lambda l, y: l.append(y) or l if y not in l else l, [(x[0], x[1]) for x in df_2.columns], []
        )
        list_no_join = list(set(col_2_train) ^ set(col_2_test))
    else:
        list_no_join = list(set(df_1.groups) ^ set(df_2.groups))
    if len(list_no_join) != 0:
        warnings.warn(f"unmatched fairness groups {list_no_join}")

    return list_no_join


def bg_red_yellow(df):
    """Color report of Class Fairness Drift based on Alert.

    Args:
        df (pd.DataFrame): report in input

    Returns:
        pd.DataFrame: color-mapping report
    """
    ret = pd.DataFrame("", index=df.index, columns=df.columns)
    for x in np.unique([x for x in df.columns.levels[0] if x != "metric"]):
        for y in np.unique(df[x].columns.get_level_values(0)):
            if "absolute_warning" in df[x][y].columns:
                ret.loc[
                    df[x][y].absolute_warning == "Red Alert", [(x, y, "curr_perf"), (x, y, "absolute_warning")]
                ] = "background-color: red"
                ret.loc[
                    df[x][y].absolute_warning == "Yellow Alert", [(x, y, "curr_perf"), (x, y, "absolute_warning")]
                ] = "background-color: yellow"

            ret.loc[
                df[x][y].relative_warning == "Red Alert", [(x, y, "drift_perc"), (x, y, "relative_warning")]
            ] = "background-color: red"
            ret.loc[
                df[x][y].relative_warning == "Yellow Alert", [(x, y, "drift_perc"), (x, y, "relative_warning")]
            ] = "background-color: yellow"
    return ret



import pandas as pd
import numpy as np
from typing import Dict
import warnings

from model_monitoring.utils import check_metrics_sets

from model_monitoring.config import read_config

params = read_config(config_dir="config", name_params="params.yml")
standard_threshold = params["performance_threshold"]


class PerformancesDrift:
    """Performance Drift Class."""

    def __init__(self, perf_metrics_curr, config_threshold=None):
        """Performance Drift Class.

        Args:
            perf_metrics_curr (dict): dictionary containing current metrics perfomances
            config_threshold (dict, optional): dictionary containing threshold settings. Defaults to None.
        """
        if not isinstance(perf_metrics_curr, Dict):
            raise ValueError(
                "Performance metrics in input has not a valid format. It should be a dictionary containing functions as keys and values as values."
            )
        if config_threshold is None:
            config_threshold = standard_threshold

        check_metrics = [
            i for i in perf_metrics_curr.keys() if (i not in config_threshold.keys()) and (i != "proportion_1")
        ]
        if len(check_metrics) > 0:
            warnings.warn(f"{check_metrics} do not have threshold settings in config_threshold")

        list_com_metrics = list(set(perf_metrics_curr.keys()).intersection(set(config_threshold.keys())))

        # initialize report
        report_df = (
            pd.DataFrame.from_dict({x: perf_metrics_curr[x] for x in list_com_metrics}, "index", columns=["curr_perf"])
            .reset_index()
            .rename(columns={"index": "metric"})
        )
        self.report = report_df
        # For binary classification we put proportion of 1 in report
        if "proportion_1" in perf_metrics_curr.keys():
            self.report.insert(1, "proportion_1_curr", perf_metrics_curr["proportion_1"])

        self.perf_metrics_curr = perf_metrics_curr
        self.config_threshold = config_threshold
        self.perf_metrics_stor = None

    def get_absolute(self):
        """Load on the report the absolute alert on current metrics perfomances."""
        # Generation Alert
        for a in self.report.metric.values:
            absolute_red = self.config_threshold[a]["absolute"]["red"]
            absolute_yellow = self.config_threshold[a]["absolute"]["yellow"]
            curr_perf = self.report.loc[self.report.metric == a, "curr_perf"].values[0]
            if self.config_threshold[a]["logic"] == "decrease":
                if absolute_red != "None":
                    if curr_perf < absolute_red:
                        self.report.loc[self.report.metric == a, "absolute_warning"] = "Red Alert"
                    else:
                        if absolute_yellow != "None":
                            if (curr_perf > absolute_red) and (curr_perf < absolute_yellow):
                                self.report.loc[self.report.metric == a, "absolute_warning"] = "Yellow Alert"
                            else:
                                self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                        else:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                else:
                    if absolute_yellow != "None":
                        if curr_perf < absolute_yellow:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = "Yellow Alert"
                        else:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                    else:
                        self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
            elif self.config_threshold[a]["logic"] == "increase":
                if absolute_red != "None":
                    if curr_perf > absolute_red:
                        self.report.loc[self.report.metric == a, "absolute_warning"] = "Red Alert"
                    else:
                        if absolute_yellow != "None":
                            if (curr_perf < absolute_red) and (curr_perf > absolute_yellow):
                                self.report.loc[self.report.metric == a, "absolute_warning"] = "Yellow Alert"
                            else:
                                self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                        else:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                else:
                    if absolute_yellow != "None":
                        if curr_perf > absolute_yellow:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = "Yellow Alert"
                        else:
                            self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
                    else:
                        self.report.loc[self.report.metric == a, "absolute_warning"] = np.nan
            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease']."
                )
        # Locate Absolute_warning after Curr_perf
        absolute_warning = self.report.pop("absolute_warning")
        self.report.insert(self.report.columns.get_loc("curr_perf") + 1, "absolute_warning", absolute_warning)

    def get_relative(self, perf_metrics_stor):
        """Load on the report the historichal metrics performances, drift compared to the current performances and relative alert on drift.

        Args:
            perf_metrics_stor (dict): dictionary containing historichal metrics perfomances.
        """
        # Check if the metrics are the same
        check_metrics_sets(metrics_1=perf_metrics_stor, metrics_2=self.perf_metrics_curr)

        list_com_metrics = [
            x
            for x in list(set(perf_metrics_stor.keys()).intersection(set(self.perf_metrics_curr.keys())))
            if x != "proportion_1"
        ]
        self.perf_metrics_stor = {x: perf_metrics_stor[x] for x in list_com_metrics}
        stor_perf_df = (
            pd.DataFrame.from_dict(self.perf_metrics_stor, "index", columns=["stor_perf"])
            .reset_index()
            .rename(columns={"index": "metric"})
        )
        # For binary classification we put proportion of 1 in report
        if "proportion_1" in perf_metrics_stor.keys():
            stor_perf_df.insert(1, "proportion_1_stor", perf_metrics_stor["proportion_1"])

        if "stor_perf" in self.report.columns:
            self.report.loc[:, "stor_perf"] = stor_perf_df.stor_perf
            if "proportion_1_stor" in stor_perf_df:
                self.report.loc[:, "proportion_1_stor"] = stor_perf_df.proportion_1_stor
        else:
            self.report = self.report.merge(stor_perf_df, how="outer", on="metric")

        # Generation Drift
        for a in self.report.metric.values:
            stor_perf = self.report.loc[self.report.metric == a, "stor_perf"].values[0]
            curr_perf = self.report.loc[self.report.metric == a, "curr_perf"].values[0]
            if self.config_threshold[a]["logic"] in ["decrease", "increase"]:
                if stor_perf > 0:
                    self.report.loc[self.report.metric == a, "drift_perc"] = (curr_perf - stor_perf) / stor_perf * 100
                else:
                    self.report.loc[self.report.metric == a, "drift_perc"] = (stor_perf - curr_perf) / stor_perf * 100
            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease']."
                )

        # Generation Alert
        for a in self.report.metric.values:
            relative_red = self.config_threshold[a]["relative"]["red"]
            relative_yellow = self.config_threshold[a]["relative"]["yellow"]
            drift_perf = self.report.loc[self.report.metric == a, "drift_perc"].values[0]
            if self.config_threshold[a]["logic"] == "decrease":
                if relative_red != "None":
                    if drift_perf < relative_red * 100:
                        self.report.loc[self.report.metric == a, "relative_warning"] = "Red Alert"
                    else:
                        if relative_yellow != "None":
                            if (drift_perf > relative_red * 100) and (drift_perf < relative_yellow * 100):
                                self.report.loc[self.report.metric == a, "relative_warning"] = "Yellow Alert"
                            else:
                                self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                        else:
                            self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                else:
                    if relative_yellow != "None":
                        if drift_perf < relative_yellow * 100:
                            self.report.loc[self.report.metric == a, "relative_warning"] = "Yellow Alert"
                        else:
                            self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                    else:
                        self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
            elif self.config_threshold[a]["logic"] == "increase":
                if relative_red != "None":
                    if drift_perf > relative_red * 100:
                        self.report.loc[self.report.metric == a, "relative_warning"] = "Red Alert"
                    else:
                        if relative_yellow != "None":
                            if (drift_perf < relative_red * 100) and (drift_perf > relative_yellow * 100):
                                self.report.loc[self.report.metric == a, "relative_warning"] = "Yellow Alert"
                            else:
                                self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                        else:
                            self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                else:
                    if relative_yellow != "None":
                        if drift_perf > relative_yellow * 100:
                            self.report.loc[self.report.metric == a, "relative_warning"] = "Yellow Alert"
                        else:
                            self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
                    else:
                        self.report.loc[self.report.metric == a, "relative_warning"] = np.nan
            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease']."
                )

    def get_report(self):
        """Return the report.

        Returns:
            pd.DataFrame: report of the class.
        """
        return self.report



import pandas as pd
import numpy as np
import warnings
from typing import Callable, List
import inspect

# classification
from model_monitoring.fairness_measures.fairness_measures import (
    statistical_parity_difference,
    disparate_impact_ratio,
    predictive_parity_difference,
    equal_opportunity_difference,
    average_odds_difference,
)

# regression
from model_monitoring.fairness_measures.fairness_measures import ddre_independence, ddre_separation, ddre_sufficiency

from model_monitoring.utils import check_size, convert_Int_dataframe, convert_Int_series
from model_monitoring.fairness_measures.fairness_measures import compute_metric_fairness


class FairnessMeasures:
    """Fairness Measures Class."""

    def __init__(self, model_type="auto", set_metrics="standard", new_metrics=None, **kwargs):
        """Fairness Measures Class.

        Args:
            model_type (str): Modelling problem among "regression", "classification", "multiclass" and "auto".
            set_metrics (str, optional): Metrics settings. It can be set "standard" for classical fairness metrics, "new" for setting a new dictionary of metrics and "add" for adding new metrics to the standard ones. Defaults to 'standard'.
            new_metrics (list, optional): list of new metrics when "set_metrics" is set to "new" and "add". Defaults to None.
        """
        # Check the model_type
        if model_type not in ["auto", "regression", "classification", "multiclass"]:
            raise ValueError(
                f"{model_type} is not a valid algo_type. It should be one of the following:\n ['auto', 'regression', 'classification', 'multiclass']"
            )
        else:
            self.model_type = model_type

        # Check the set_metrics
        if set_metrics not in ["standard", "add", "new"]:
            raise ValueError(
                f"{set_metrics} is not a valid set_metrics. It should be one of the following:\n ['standard', 'add', 'new']"
            )
        self.set_metrics = set_metrics

        # Check new_metrics
        if self.set_metrics in ["add", "new"]:
            if new_metrics is None:
                self.new_metrics = []
            else:
                if isinstance(new_metrics, List):
                    self.new_metrics = new_metrics
                elif isinstance(new_metrics, Callable):
                    self.new_metrics = [new_metrics]
                else:
                    raise ValueError(f"{new_metrics} has not a valid format. It should be a list containing functions")

        # Set the metrics for each set_metric case
        if self.set_metrics == "new":
            self.metrics = self.new_metrics
        if self.set_metrics in ["standard", "add"]:
            if self.model_type == "regression":
                self.metrics = [ddre_independence, ddre_separation, ddre_sufficiency]
            elif self.model_type == "classification":
                self.metrics = [
                    statistical_parity_difference,
                    disparate_impact_ratio,
                    predictive_parity_difference,
                    equal_opportunity_difference,
                    average_odds_difference,
                ]
            else:
                self.metrics = [
                    statistical_parity_difference,
                    predictive_parity_difference,
                    equal_opportunity_difference,
                    average_odds_difference,
                ]
            if self.set_metrics == "add":
                self.metrics = self.metrics + self.new_metrics

        self.target = None
        self.predictions = None
        self.X_fair = None
        self.fair_feat = None
        self.perf_metrics = None

        for k, v in kwargs.items():
            self.__dict__[k] = v

    def compute_metrics(self, target, predictions, X_fair, fair_feat=None, return_prop_true=True, **kwargs):
        """Compute fairness metrics.

        Args:
            target (np.array/pd.Series): target column
            predictions (np.array/pd.Series): Predictions array.
            X_fair (pd.DataFrame/pd.Series): data to be analyzed for fairness metric.
            fair_feat (str/list, optional): list of features (or list of features) to be analyzed for fairness metric. Defaults to None.
            return_prop_true (bool, optional): boolean that determines whether to return the portion of the target in binary classification. Defaults to None.

        Returns:
            dict: fairness metrics performances
        """
        for k, v in kwargs.items():
            self.__dict__[k] = v

        # Check size of the predictions and target
        check_size(predictions, target)

        # Check size of the target and the features to be analyzed
        check_size(target, X_fair)

        if isinstance(target, np.ndarray):
            target = pd.Series(target, name="target")
        vals = target.nunique()
        if vals == 1:
            warnings.warn("The target column selected is constant")
        elif self.model_type == "auto":
            if vals <= 2:
                self.model_type = "classification"
            elif vals < 11:
                self.model_type = "multiclass"
            else:
                self.model_type = "regression"

        if isinstance(predictions, np.ndarray):
            predictions = pd.Series(predictions, name="predictions")
        if isinstance(X_fair, pd.Series):
            X_fair = pd.DataFrame(X_fair, index=X_fair.index, columns=[X_fair.name])
        if isinstance(fair_feat, str):
            fair_feat = [fair_feat]

        if fair_feat is None:
            fair_feat = X_fair.columns

        self.target = convert_Int_series(target)
        self.predictions = convert_Int_series(predictions)
        self.X_fair = convert_Int_dataframe(X_fair)
        self.fair_feat = fair_feat

        self.return_prop_true = return_prop_true
        perf_metrics = dict()
        if self.model_type in ["regression", "classification", "multiclass"]:
            for i in self.metrics:
                dict_feat = dict()
                for z in self.fair_feat:
                    dict_to_use = {
                        k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()
                    }
                    dict_labels = dict()
                    for j in self.X_fair[z].dropna().drop_duplicates().values:
                        if isinstance(j, List) or isinstance(j, np.ndarray):
                            j_str = ", ".join(map(str, j))
                            perc_label = (
                                self.X_fair[z][(self.X_fair[z] == j).all(axis=1)].shape[0] / self.X_fair[z].shape[0]
                            )
                        else:
                            j_str = j
                            perc_label = self.X_fair[z][self.X_fair[z] == j].shape[0] / self.X_fair[z].shape[0]
                        dict_labels[j_str] = [
                            compute_metric_fairness(
                                self.target,
                                self.predictions,
                                metric=i,
                                fair_attr=self.X_fair[z],
                                unpriv_group=j,
                                **dict_to_use,
                            ),
                            perc_label,
                        ]
                        if self.model_type == "classification":
                            if self.return_prop_true:
                                try:  # One Fairness Attribute
                                    dict_labels[j_str].append(
                                        (self.target[self.X_fair[z] == j] == 1).mean()
                                    )  # Add proportion of 1 label
                                except Exception:  # More Fairness Attributes
                                    dict_labels[j_str].append(
                                        (self.target[(self.X_fair[z] == j).all(axis=1)] == 1).mean()
                                    )  # Add proportion of 1 label
                    if isinstance(z, List):
                        z_str = ", ".join(map(str, z))
                    else:
                        z_str = z
                    dict_feat[z_str] = dict_labels
                perf_metrics[i.__name__] = dict_feat
        else:
            raise ValueError("Invalid model type.")

        self.perf_metrics = perf_metrics
        return self.perf_metrics



# regression
from sklearn.metrics import (
    mean_squared_error,
    r2_score,
    explained_variance_score,
    median_absolute_error,
)

# classification
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    roc_auc_score,
    balanced_accuracy_score,
)

import warnings
import pandas as pd
import numpy as np
from typing import Dict
import inspect

from model_monitoring.utils import check_size, convert_Int_series
from model_monitoring.performance_measures.performance_measures import compute_metric


class PerformancesMeasures:
    """Performance Measures Class."""

    def __init__(self, model_type="auto", set_metrics="standard", new_metrics=None, **kwargs):
        """Performance Measures Class.

        Args:
            model_type (str): Modelling problem among "regression", "classification", "multiclass" and "auto".
            set_metrics (str, optional): Metrics settings. It can be set "standard" for classical ML metrics, "new" for setting a new dictionary of metrics and "add" for adding new metrics to the standard ones. Defaults to 'standard'.
            new_metrics (dict, optional): Dictionary of new metrics as keys when "set_metrics" is set to "new" and "add", and one among ['pred','prob'] as values. Defaults to None.
        """
        # Check the model_type
        if model_type not in ["auto", "regression", "classification", "multiclass"]:
            raise ValueError(
                f"{model_type} is not a valid algo_type. It should be one of the following:\n ['auto', 'regression', 'classification', 'multiclass']"
            )
        else:
            self.model_type = model_type

        # Check the set_metrics
        if set_metrics not in ["standard", "add", "new"]:
            raise ValueError(
                f"{set_metrics} is not a valid set_metrics. It should be one of the following:\n ['standard', 'add', 'new']"
            )
        self.set_metrics = set_metrics

        # Check new_metrics
        if self.set_metrics in ["add", "new"]:
            if new_metrics is None:
                self.new_metrics = {}
            else:
                if isinstance(new_metrics, Dict):
                    if set(new_metrics.values()).issubset(set(["pred", "prob"])):
                        self.new_metrics = new_metrics
                    else:
                        raise ValueError(
                            f"{list(set(new_metrics.values()))} contains invalid input. Valid inputs are ['pred', 'prob']"
                        )
                else:
                    raise ValueError(
                        f"{new_metrics} has not a valid format. It should be a dictionary containing functions as keys and one of ['pred', 'prob'] as values."
                    )

        # Set the metrics for each set_metric case
        if self.set_metrics == "new":
            self.metrics = self.new_metrics
        if self.set_metrics in ["standard", "add"]:
            if self.model_type == "regression":
                self.metrics = {
                    mean_squared_error: "pred",
                    r2_score: "pred",
                    median_absolute_error: "pred",
                    explained_variance_score: "pred",
                }
            elif self.model_type == "classification":
                self.metrics = {
                    balanced_accuracy_score: "pred",
                    accuracy_score: "pred",
                    precision_score: "pred",
                    recall_score: "pred",
                    f1_score: "pred",
                    roc_auc_score: "prob",
                }
            else:
                self.metrics = {balanced_accuracy_score: "pred", accuracy_score: "pred", roc_auc_score: "prob"}
            if self.set_metrics == "add":
                self.metrics = {**self.metrics, **self.new_metrics}

        self.predictions = None
        self.prob = None
        self.target = None
        self.perf_metrics = None
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def compute_metrics(self, target, predictions=None, prob=None, return_prop_true=True, **kwargs):
        """Compute metrics performances.

        Args:
            target (np.array/pd.Series): target column
            predictions (np.array/pd.Series, optional): Predictions array. Defaults to None.
            prob (np.array/pd.Series, optional): Probabilities prob for classification. Defaults to None.
            return_prop_true (bool, optional): boolean that determines whether to return the portion of the target in binary classification. Defaults to None.

        Returns:
            dict: metrics performances
        """
        for k, v in kwargs.items():
            self.__dict__[k] = v

        # Check one among predictions and prob exists
        if (predictions is None) and (prob is None):
            raise ValueError("at least one among predictions and prob must be not None")

        # Check size of the predictions and target
        if predictions is not None:
            check_size(predictions, target)

        # Check size of the target and prob
        if prob is not None:
            check_size(target, prob)

        if isinstance(target, np.ndarray):
            target = pd.Series(target, name="target")
        vals = target.nunique()
        if vals == 1:
            warnings.warn("The target column selected is constant")
        elif self.model_type == "auto":
            if vals <= 2:
                self.model_type = "classification"
            elif vals < 11:
                self.model_type = "multiclass"
            else:
                self.model_type = "regression"

        if self.model_type == "regression":
            if predictions is None:
                raise ValueError("predictions not provided")

        self.target = convert_Int_series(target)
        if predictions is not None:
            self.predictions = convert_Int_series(predictions)
        else:
            self.predictions = predictions
        if prob is not None:
            self.prob = convert_Int_series(prob)
        else:
            self.prob = prob

        self.return_prop_true = return_prop_true
        perf_metrics = dict()
        if self.model_type == "regression":
            for i, j in self.metrics.items():
                dict_to_use = {k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()}
                if j == "pred":
                    perf_metrics[i.__name__] = compute_metric(self.target, self.predictions, metric=i, **dict_to_use)
                else:
                    warnings.warn(f"{j} is a wrong label for regression model type. Label {i.__name__} with 'pred'.")
        elif self.model_type == "classification":
            if self.predictions is None:
                try:
                    self.predictions = self.prob.apply(lambda x: 1 if x > 0.5 else 0)
                except Exception:
                    self.predictions = np.array([1 if i > 0.5 else 0 for i in self.prob])
            for i, j in self.metrics.items():
                dict_to_use = {k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()}
                # some metrics don't work well with 1 target value (e.g. roc_auc_score)
                if vals == 1:
                    if i.__name__ in ["roc_auc_score", "lift_score", "gain_score"]:
                        warnings.warn(f"{i.__name__} cannot be used when target has a constant value.")
                        continue
                if j == "prob":
                    if self.prob is None:
                        warnings.warn(f"{i.__name__} needs prob, but prob are not provided")
                    else:
                        perf_metrics[i.__name__] = compute_metric(self.target, self.prob, metric=i, **dict_to_use)
                else:
                    perf_metrics[i.__name__] = compute_metric(self.target, self.predictions, metric=i, **dict_to_use)
            if self.return_prop_true:
                perf_metrics["proportion_1"] = (self.target == 1).mean()  # Add proportion of 1 label
        elif self.model_type == "multiclass":
            if self.predictions is None:
                self.predictions = pd.Series([np.argmax(x) for x in self.prob], name="prob")
            for i, j in self.metrics.items():
                dict_to_use = {k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()}
                if i.__name__ in [
                    "precision_score",
                    "recall_score",
                    "f1_score",
                    "fbeta_score",
                    "brier_score_loss",
                    "class_likelihood_ratios",
                    "dcg_score",
                    "jaccard_score",
                    "log_loss",
                    "matthews_corrcoef",
                    "ndcg_score",
                ]:
                    warnings.warn(f"{i.__name__} is used for binary classification")
                elif j == "prob":
                    if self.prob is None:
                        warnings.warn(f"{i.__name__} needs prob, but prob are not provided")
                    else:
                        perf_metrics[i.__name__] = compute_metric(self.target, self.prob, metric=i, multi_class="ovr")
                else:
                    perf_metrics[i.__name__] = compute_metric(self.target, self.predictions, metric=i, **dict_to_use)
        else:
            raise ValueError("Invalid model type.")

        self.perf_metrics = perf_metrics
        return self.perf_metrics



from warnings import warn
import numpy as np

from sklearn.metrics import recall_score, precision_score
from sklearn.metrics import confusion_matrix
from sklearn.linear_model import LogisticRegression


class UndefinedMetricWarning(UserWarning):
    """Warning used when the metric is invalid."""


# ============================ GROUP FAIRNESS CLASSIFICATION ==================================
def statistical_parity_difference(y_true, y_pred, fair_attr, unpriv_group=0, multiclass_agg_mode="max", **kwargs):
    """Fairness metric for binary classification: it returns the difference between the computed PR (Positive Rate) between the two groups.

    Args:
        y_true (pd.Series/np.array): ground truth target values.
        y_pred (pd.Series/np.array): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.
        multiclass_agg_mode (str, optional): mode of aggregating metric scores for multiclassification problem among 'max' and 'sum'. Defaults to 'max'.

    Returns:
        float: statistical parity difference metric score
    """
    multiclass = False
    # For multiclassifcation ML problem
    if len(np.unique(y_true)) > 2:
        if multiclass_agg_mode not in ["max", "sum"]:
            raise ValueError(f"{multiclass_agg_mode} is not a valid aggregation mode. Choose between ['max','sum'].")
        list_score = []
        for cl in np.unique(y_true):
            y_true_multi = np.where(y_true == cl, 1, 0)
            y_pred_multi = np.where(y_pred == cl, 1, 0)
            # One Fairness Attribute
            try:
                unpr_pr_group = positive_rate(
                    y_true_multi[fair_attr == unpriv_group], y_pred_multi[fair_attr == unpriv_group], **kwargs
                )
                pr_pr_group = positive_rate(
                    y_true_multi[fair_attr != unpriv_group], y_pred_multi[fair_attr != unpriv_group], **kwargs
                )
            # More Fairness Attributes
            except Exception:
                unpr_pr_group = positive_rate(
                    y_true_multi[(fair_attr == unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr == unpriv_group).all(axis=1)],
                    **kwargs,
                )
                pr_pr_group = positive_rate(
                    y_true_multi[(fair_attr != unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr != unpriv_group).all(axis=1)],
                    **kwargs,
                )
            list_score.append(unpr_pr_group - pr_pr_group)
        multiclass = True
    else:
        # One Fairness Attribute
        try:
            unpr_pr_group = positive_rate(
                y_true[fair_attr == unpriv_group], y_pred[fair_attr == unpriv_group], **kwargs
            )
            pr_pr_group = positive_rate(y_true[fair_attr != unpriv_group], y_pred[fair_attr != unpriv_group], **kwargs)
        # More Fairness Attributes
        except Exception:
            unpr_pr_group = positive_rate(
                y_true[(fair_attr == unpriv_group).all(axis=1)],
                y_pred[(fair_attr == unpriv_group).all(axis=1)],
                **kwargs,
            )
            pr_pr_group = positive_rate(
                y_true[(fair_attr != unpriv_group).all(axis=1)],
                y_pred[(fair_attr != unpriv_group).all(axis=1)],
                **kwargs,
            )
    if multiclass:
        if multiclass_agg_mode == "max":
            return max(list_score, key=abs)
        else:
            return sum(map(abs, list_score))
    else:
        return unpr_pr_group - pr_pr_group


def disparate_impact_ratio(y_true, y_pred, fair_attr, unpriv_group=0, **kwargs):
    """Fairness metric for binary classification: it returns the ratio between the computed PR (Positive Rate) between the two groups.

    Args:
        y_true (pd.Series/np.array): ground truth target values.
        y_pred (pd.Series/np.array): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.

    Returns:
        float: disparate impact ratio metric score
    """
    # One Fairness Attribute
    try:
        unpr_pr_group = positive_rate(y_true[fair_attr == unpriv_group], y_pred[fair_attr == unpriv_group], **kwargs)
        pr_pr_group = positive_rate(y_true[fair_attr != unpriv_group], y_pred[fair_attr != unpriv_group], **kwargs)
    # More Fairness Attributes
    except Exception:
        unpr_pr_group = positive_rate(
            y_true[(fair_attr == unpriv_group).all(axis=1)], y_pred[(fair_attr == unpriv_group).all(axis=1)], **kwargs
        )
        pr_pr_group = positive_rate(
            y_true[(fair_attr != unpriv_group).all(axis=1)], y_pred[(fair_attr != unpriv_group).all(axis=1)], **kwargs
        )

    return unpr_pr_group / pr_pr_group


def predictive_parity_difference(
    y_true, y_pred, fair_attr, unpriv_group=0, multiclass_agg_mode="max", zero_division=0, **kwargs
):
    """Fairness metric for binary classification: it returns the difference between the computed precision between the two groups.

    Args:
        y_true (pd.Series/np.array): ground truth target values.
        y_pred (pd.Series/np.array): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.
        multiclass_agg_mode (str, optional): mode of aggregating metric scores for multiclassification problem among 'max' and 'sum'. Defaults to 'max'.
        zero_division (int/str optional): it can be 0, 1 or 'warn'. It sets to a value when division with 0 in the denominator.
            When 'warn' it prints the warning and set the division to 0. Deafults to 0.

    Returns:
        float: predictive parity difference metric score
    """
    multiclass = False
    # For multiclassifcation ML problem
    if len(np.unique(y_true)) > 2:
        if multiclass_agg_mode not in ["max", "sum"]:
            raise ValueError(f"{multiclass_agg_mode} is not a valid aggregation mode. Choose between ['max','sum'].")
        list_score = []
        for cl in np.unique(y_true):
            y_true_multi = np.where(y_true == cl, 1, 0)
            y_pred_multi = np.where(y_pred == cl, 1, 0)
            # One Fairness Attribute
            try:
                unpr_pre_group = precision_score(
                    y_true_multi[fair_attr == unpriv_group],
                    y_pred_multi[fair_attr == unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_pre_group = precision_score(
                    y_true_multi[fair_attr != unpriv_group],
                    y_pred_multi[fair_attr != unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
            # More Fairness Attributes
            except Exception:
                unpr_pre_group = precision_score(
                    y_true_multi[(fair_attr == unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr == unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_pre_group = precision_score(
                    y_true_multi[(fair_attr != unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr != unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
            list_score.append(unpr_pre_group - pr_pre_group)
        multiclass = True
    else:
        # One Fairness Attribute
        try:
            unpr_pre_group = precision_score(
                y_true[fair_attr == unpriv_group],
                y_pred[fair_attr == unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
            pr_pre_group = precision_score(
                y_true[fair_attr != unpriv_group],
                y_pred[fair_attr != unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
        # More Fairness Attributes
        except Exception:
            unpr_pre_group = precision_score(
                y_true[(fair_attr == unpriv_group).all(axis=1)],
                y_pred[(fair_attr == unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
            pr_pre_group = precision_score(
                y_true[(fair_attr != unpriv_group).all(axis=1)],
                y_pred[(fair_attr != unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
    if multiclass:
        if multiclass_agg_mode == "max":
            return max(list_score, key=abs)
        else:
            return sum(map(abs, list_score))
    else:
        return unpr_pre_group - pr_pre_group


def equal_opportunity_difference(
    y_true, y_pred, fair_attr, unpriv_group=0, multiclass_agg_mode="max", zero_division=0, **kwargs
):
    """Fairness metric for binary classification: it returns the difference between the computed recall between two groups.

    Args:
        y_true (pd.Series/np.array): ground truth target values.
        y_pred (pd.Series/np.array): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.
        multiclass_agg_mode (str, optional): mode of aggregating metric scores for multiclassification problem among 'max' and 'sum'. Defaults to 'max'.
        zero_division (int/str optional): it can be 0, 1 or 'warn'. It sets to a value when division with 0 in the denominator.
            When 'warn' it prints the warning and set the division to 0. Deafults to 0.

    Returns:
        float: equal opportunity difference metric score
    """
    multiclass = False
    # For multiclassifcation ML problem
    if len(np.unique(y_true)) > 2:
        if multiclass_agg_mode not in ["max", "sum"]:
            raise ValueError(f"{multiclass_agg_mode} is not a valid aggregation mode. Choose between ['max','sum'].")
        list_score = []
        for cl in np.unique(y_true):
            y_true_multi = np.where(y_true == cl, 1, 0)
            y_pred_multi = np.where(y_pred == cl, 1, 0)
            # One Fairness Attribute
            try:
                unpr_rec_group = recall_score(
                    y_true_multi[fair_attr == unpriv_group],
                    y_pred_multi[fair_attr == unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_rec_group = recall_score(
                    y_true_multi[fair_attr != unpriv_group],
                    y_pred_multi[fair_attr != unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
            # More Fairness Attributes
            except Exception:
                unpr_rec_group = recall_score(
                    y_true_multi[(fair_attr == unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr == unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_rec_group = recall_score(
                    y_true_multi[(fair_attr != unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr != unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
            list_score.append(unpr_rec_group - pr_rec_group)
        multiclass = True
    else:
        # One Fairness Attribute
        try:
            unpr_rec_group = recall_score(
                y_true[fair_attr == unpriv_group],
                y_pred[fair_attr == unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
            pr_rec_group = recall_score(
                y_true[fair_attr != unpriv_group],
                y_pred[fair_attr != unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
        # More Fairness Attributes
        except Exception:
            unpr_rec_group = recall_score(
                y_true[(fair_attr == unpriv_group).all(axis=1)],
                y_pred[(fair_attr == unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
            pr_rec_group = recall_score(
                y_true[(fair_attr != unpriv_group).all(axis=1)],
                y_pred[(fair_attr != unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
    if multiclass:
        if multiclass_agg_mode == "max":
            return max(list_score, key=abs)
        else:
            return sum(map(abs, list_score))
    else:
        return unpr_rec_group - pr_rec_group


def average_odds_difference(
    y_true, y_pred, fair_attr, unpriv_group=0, multiclass_agg_mode="max", zero_division=0, **kwargs
):
    """Fairness metric for binary classification: it returns the average of the differences between the computed recall and FPR (False Positive Rate) between two groups.

    Args:
        y_true (pd.Series/np.array): ground truth target values.
        y_pred (pd.Series/np.array): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.
        multiclass_agg_mode (str, optional): mode of aggregating metric scores for multiclassification problem among 'max' and 'sum'. Defaults to 'max'.
        zero_division (int/str optional): it can be 0, 1 or 'warn'. It sets to a value when division with 0 in the denominator.
            When 'warn' it prints the warning and set the division to 0. Deafults to 0.

    Returns:
        float: average odd difference metric score
    """
    multiclass = False
    # For multiclassifcation ML problem
    if len(np.unique(y_true)) > 2:
        if multiclass_agg_mode not in ["max", "sum"]:
            raise ValueError(f"{multiclass_agg_mode} is not a valid aggregation mode. Choose between ['max','sum'].")
        list_score = []
        for cl in np.unique(y_true):
            y_true_multi = np.where(y_true == cl, 1, 0)
            y_pred_multi = np.where(y_pred == cl, 1, 0)
            # One Fairness Attribute
            try:
                unpr_fpr_group = false_positive_rate(
                    y_true_multi[fair_attr == unpriv_group],
                    y_pred_multi[fair_attr == unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_fpr_group = false_positive_rate(
                    y_true_multi[fair_attr != unpriv_group],
                    y_pred_multi[fair_attr != unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
                unpr_rec_group = recall_score(
                    y_true_multi[fair_attr == unpriv_group],
                    y_pred_multi[fair_attr == unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_rec_group = recall_score(
                    y_true_multi[fair_attr != unpriv_group],
                    y_pred_multi[fair_attr != unpriv_group],
                    zero_division=zero_division,
                    **kwargs,
                )
            # More Fairness Attributes
            except Exception:
                unpr_fpr_group = false_positive_rate(
                    y_true_multi[(fair_attr == unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr == unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_fpr_group = false_positive_rate(
                    y_true_multi[(fair_attr != unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr != unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
                unpr_rec_group = recall_score(
                    y_true_multi[(fair_attr == unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr == unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
                pr_rec_group = recall_score(
                    y_true_multi[(fair_attr != unpriv_group).all(axis=1)],
                    y_pred_multi[(fair_attr != unpriv_group).all(axis=1)],
                    zero_division=zero_division,
                    **kwargs,
                )
            list_score.append(((unpr_fpr_group - pr_fpr_group) + (unpr_rec_group - pr_rec_group)) / 2)
        multiclass = True
    else:
        # One Fairness Attribute
        try:
            unpr_fpr_group = false_positive_rate(
                y_true[fair_attr == unpriv_group],
                y_pred[fair_attr == unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
            pr_fpr_group = false_positive_rate(
                y_true[fair_attr != unpriv_group],
                y_pred[fair_attr != unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
            unpr_rec_group = recall_score(
                y_true[fair_attr == unpriv_group],
                y_pred[fair_attr == unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
            pr_rec_group = recall_score(
                y_true[fair_attr != unpriv_group],
                y_pred[fair_attr != unpriv_group],
                zero_division=zero_division,
                **kwargs,
            )
        # More Fairness Attributes
        except Exception:
            unpr_fpr_group = false_positive_rate(
                y_true[(fair_attr == unpriv_group).all(axis=1)],
                y_pred[(fair_attr == unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
            pr_fpr_group = false_positive_rate(
                y_true[(fair_attr != unpriv_group).all(axis=1)],
                y_pred[(fair_attr != unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
            unpr_rec_group = recall_score(
                y_true[(fair_attr == unpriv_group).all(axis=1)],
                y_pred[(fair_attr == unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
            pr_rec_group = recall_score(
                y_true[(fair_attr != unpriv_group).all(axis=1)],
                y_pred[(fair_attr != unpriv_group).all(axis=1)],
                zero_division=zero_division,
                **kwargs,
            )
    if multiclass:
        if multiclass_agg_mode == "max":
            return max(list_score, key=abs)
        else:
            return sum(map(abs, list_score))
    else:
        return ((unpr_fpr_group - pr_fpr_group) + (unpr_rec_group - pr_rec_group)) / 2


# ============================ GROUP FAIRNESS REGRESSION ==================================


def ddre_independence(
    y_true,
    y_pred,
    fair_attr,
    unpriv_group=0,
):
    """It computes the Direct Density Ratio Estimation (DDRE) of the independence density ratio score.

    Args:
        y_true (pd.Series): ground truth target values (not used for the metrics). It is set for compatibility.
        y_pred (pd.Series): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.

    Returns:
        float: independence DDRE score
    """
    # normalization
    s_u = ((y_pred - y_pred.mean()) / y_pred.std()).values.reshape(-1, 1)

    # One Fairness Attribute
    if (len(fair_attr.shape) == 1) | ((fair_attr.shape[-1]) == 1):
        a = np.where(fair_attr == unpriv_group, 1, 0)
    # More Fairness Attributes
    else:
        a = np.where((fair_attr == unpriv_group).all(axis=1), 1, 0)

    n = len(a)

    # density ratio approximation with probabilistic classifier
    p_s = LogisticRegression()
    p_s.fit(s_u, a.ravel())
    pred_p_s = p_s.predict_proba(s_u.reshape(-1, 1))[:, 1]

    # independence metric score
    r_ind = ((n - a.sum()) / a.sum()) * (pred_p_s / (1 - pred_p_s)).mean()

    return r_ind


def ddre_separation(y_true, y_pred, fair_attr, unpriv_group=0):
    """It computes the Direct Density Ratio Estimation (DDRE) of the separation density ratio score.

    Args:
        y_true (pd.Series): ground truth target values.
        y_pred (pd.Series): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.

    Returns:
        float: separation DDRE score
    """
    # normalization
    y_u = ((y_true - y_true.mean()) / y_true.std()).values.reshape(-1, 1)
    s_u = ((y_pred - y_pred.mean()) / y_pred.std()).values.reshape(-1, 1)

    # One Fairness Attribute
    if (len(fair_attr.shape) == 1) | ((fair_attr.shape[-1]) == 1):
        a = np.where(fair_attr == unpriv_group, 1, 0)
    # More Fairness Attributes
    else:
        a = np.where((fair_attr == unpriv_group).all(axis=1), 1, 0)

    # density ratios approximation with probabilistic classifier
    p_ys = LogisticRegression()
    p_y = LogisticRegression()
    p_y.fit(y_u, a.ravel())
    p_ys.fit(np.c_[y_u, s_u], a.ravel())
    pred_p_y = p_y.predict_proba(y_u.reshape(-1, 1))[:, 1]
    pred_p_ys = p_ys.predict_proba(np.c_[y_u, s_u])[:, 1]

    # separation metric score
    r_sep = ((pred_p_ys / (1 - pred_p_ys) * (1 - pred_p_y) / pred_p_y)).mean()

    return r_sep


def ddre_sufficiency(y_true, y_pred, fair_attr, unpriv_group=0):
    """It computes the Direct Density Ratio Estimation (DDRE) of the sufficiency density ratio score.

    Args:
        y_true (pd.Series): ground truth target values.
        y_pred (pd.Series): estimated targets.
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics.
        unpriv_group (int/float/str/list, optional): the label of the unprivileged group in the features fair_attr. Defaults to 0.

    Returns:
        float: sufficiency DDRE score
    """
    # normalization
    y_u = ((y_true - y_true.mean()) / y_true.std()).values.reshape(-1, 1)
    s_u = ((y_pred - y_pred.mean()) / y_pred.std()).values.reshape(-1, 1)

    # One Fairness Attribute
    if (len(fair_attr.shape) == 1) | ((fair_attr.shape[-1]) == 1):
        a = np.where(fair_attr == unpriv_group, 1, 0)
    # More Fairness Attributes
    else:
        a = np.where((fair_attr == unpriv_group).all(axis=1), 1, 0)

    # density ratios approximation with probabilistic classifier
    p_s = LogisticRegression()
    p_ys = LogisticRegression()
    p_s.fit(s_u, a.ravel())
    p_ys.fit(np.c_[y_u, s_u], a.ravel())
    pred_p_s = p_s.predict_proba(s_u.reshape(-1, 1))[:, 1]
    pred_p_ys = p_ys.predict_proba(np.c_[y_u, s_u])[:, 1]

    # sufficiency metric score
    r_suf = ((pred_p_ys / (1 - pred_p_ys)) * ((1 - pred_p_s) / pred_p_s)).mean()

    return r_suf


# ================================ USEFUL METRICS =========================================


def specificity_score(y_true, y_pred, zero_division=0, **kwargs):
    """Retrieve the specificity metric for a binary classification ML problem.

    Args:
        y_true (pandas.Series/np.array): ground truth target values.
        y_pred (pandas.Series/np.array): estimated targets.
        zero_division (int/str optional): it can be 0, 1 or 'warn'. It sets to a value when division with 0 in the denominator.
            When 'warn' it prints the warning and set the division to 0. Deafults to 0.

    Returns:
        float: specificity metric score
    """
    try:
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1], **kwargs).ravel()
    except Exception:
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[-1, 1], **kwargs).ravel()

    if (tn == 0) and (fp == 0):
        if zero_division not in ["warn", 0, 1]:
            raise ValueError(
                f"{zero_division} not valid logic for division with 0 in the denominator. It should be one of the following : ['warn',0,1]"
            )
        elif zero_division == "warn":
            warn("Specificity is ill-defined and being set to 0.0 due to no false samples.", UndefinedMetricWarning)
            specificity = 0
        else:
            specificity = zero_division
    else:
        specificity = tn / (tn + fp)
    return specificity


def positive_rate(y_true, y_pred, **kwargs):
    """Retrieve the PR (Positive Rate) metric for a binary classification ML problem.

    Args:
        y_true (pandas.Series/np.array): ground truth target values.
        y_pred (pandas.Series/np.array): estimated targets.

    Returns:
        float: PR metric score
    """
    try:
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1], **kwargs).ravel()
    except Exception:
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[-1, 1], **kwargs).ravel()

    pr = (tp + fp) / len(y_true)

    return pr


def false_positive_rate(y_true, y_pred, zero_division=0, **kwargs):
    """Retrieve the FPR (False Positive Rate) for a binary classification ML problem.

    Args:
        y_true (pandas.Series/np.array): ground truth target values.
        y_pred (pandas.Series/np.array): estimated targets.
        zero_division (int/str optional): it can be 0, 1 or 'warn'. It sets to a value when division with 0 in the denominator.
            When 'warn' it prints the warning and set the division to 0. Deafults to 0.

    Returns:
        float: FPR metric score
    """
    return 1 - specificity_score(y_true, y_pred, zero_division=zero_division, **kwargs)


# ================================ USEFUL FUNCTION =========================================


def compute_metric_fairness(y_true, y_pred, metric, fair_attr, unpriv_group, **kwargs):
    """Compute the supervised fairness metric scoring.

    Args:
        y_pred (pd.Series/np.array): predictions/scores of target
        metric (function): fairness metric function
        fair_attr (pd.DataFrame/pd.Series): attributes to be analyzed for fairness metrics
        unpriv_group (int/float/str/list): the label of the unprivileged group in the features fair_attr
        y_true (pd.Series/np.array, optional): true values of target. Defaults to None.

    Returns:
        fairness metric scoring performance
    """
    return metric(y_true, y_pred, fair_attr, unpriv_group, **kwargs)



import pandas as pd

from model_monitoring.utils import (
    get_categorical_features,
    retrieve_bin_numerical,
    merge_numerical_bins,
    merge_categorical_bins,
)


def retrieve_bins_dict(df, features, nbins=1000, bin_min_pct=0.04):
    """Retrieve metadata bins mapping for all the features.

    Args:
        df (pd.DataFrame): input dataframe.
        features (list): list of features to be checked.
        nbins (int, optional): number of bins into which the features will be bucketed (maximum). Defaults to 1000.
        bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.

    Returns:
        dict: dictionary reporting for each feature the metadata bin mapping
    """
    if len(features) == 0:
        features = df.columns
    if isinstance(features, str):
        features = [features]

    bins_dict = dict()

    for col in features:
        bin_col = dict()
        # categorical bins generation
        if col in get_categorical_features(df):
            mapper = {x: x for x in df[col].dropna().unique()}
            mapper = merge_categorical_bins(df, col, mapper, bin_min_pct=bin_min_pct)
            base_bin = df[col].dropna().map(mapper)
            for y in list(dict.fromkeys(mapper.values())):
                bin_col[y] = {
                    "labels": [k for k, v in mapper.items() if v == y],
                    "freq": base_bin.value_counts(normalize=True)[y],
                }
        # numerical bins generation
        else:
            cuts = retrieve_bin_numerical(df, col, max_n_bins=nbins)
            cuts = merge_numerical_bins(df, col, cuts, bin_min_pct=bin_min_pct)
            base_bin = pd.cut(df[col].dropna(), cuts, right=True, precision=25, duplicates="drop")
            for i, y in enumerate(base_bin.value_counts(normalize=True).sort_index().index):
                bin_col[f"bin_{i}"] = {"min": y.left, "max": y.right, "freq": base_bin.value_counts(normalize=True)[y]}

        bins_dict[col] = bin_col

    return bins_dict


def map_bins_dict(df, meta_dict, features):
    """Apply metadata mapping to an input dataset.

    Args:
        df (pd.DataFrame): input dataframe.
        meta_dict (dict): reference metadata dictionary.
        features (list): list of features to be checked.

    Returns:
        dict: dictionary reporting for each feature the metadata bin mapping
    """
    if len(features) == 0:
        features = df.columns
    if isinstance(features, str):
        features = [features]

    bins_dict = dict()

    for col in features:
        bin_col = dict()

        # numerical bins generation
        if meta_dict[col]["type"] == "numerical":
            for x in [
                y
                for y in meta_dict[col].keys()
                if y not in ["type", "min_val", "max_val", "missing_values", "not_missing_values"]
            ]:
                min_bin = meta_dict[col][x]["min"]
                max_bin = meta_dict[col][x]["max"]
                bin_col.update(
                    {
                        x: {
                            "min": min_bin,
                            "max": max_bin,
                            "freq": df.loc[(df[col] > min_bin) & (df[col] <= max_bin), col].dropna().shape[0]
                            / df.dropna().shape[0],
                        }
                    }
                )

        # categorical bins generation
        if meta_dict[col]["type"] == "categorical":
            for x in [y for y in meta_dict[col].keys() if y not in ["type", "missing_values", "not_missing_values"]]:
                list_new_labels = [k for k in meta_dict[col][x]["labels"] if k in df[col].dropna().unique()]
                if list_new_labels != []:
                    bin_col.update(
                        {
                            x: {
                                "labels": list_new_labels,
                                "freq": df[col].dropna().value_counts(normalize=True)[list_new_labels].sum(),
                            }
                        }
                    )
            new_data = set(df[col].dropna().unique()) - set(sum([bin_col[x]["labels"] for x in bin_col.keys()], []))
            if new_data != set():
                bin_col.update(
                    {
                        "_other_": {
                            "labels": list(new_data),
                            "freq": df[col].dropna().value_counts(normalize=True)[list(new_data)].sum(),
                        }
                    }
                )

        bins_dict[col] = bin_col

    return bins_dict



import yaml
import pkgutil

def read_config(config_dir="config", name_params="drift_threshold.yml"):
    """
    Return the project configuration settings.

    Args:
        config_dir (str): Name of directory containing config
        name_params (str): name of the file yaml containing configurations

    Returns:
        (dict): combined general settings with params settings
    """
    params_config_path = pkgutil.get_data(__name__, f"{config_dir}/{name_params}")
    configurations = yaml.safe_load(params_config_path)

    return configurations


import numpy as np
import pandas as pd
from tqdm import tqdm
import sys
import warnings
from scipy import stats

from model_monitoring.utils import (
    get_categorical_features,
    retrieve_bin_numerical,
    merge_numerical_bins,
    merge_categorical_bins,
)


def stat_report(
    base_df,
    compare_df,
    type_data="data",
    feat_to_check=None,
    stat="psi",
    alpha=0.05,
    max_psi=0.2,
    mid_psi=0.1,
    psi_nbins=1000,
    psi_bin_min_pct=0.04,
    return_meta_ref=False,
    dim_threshold=5000,
):
    """Retrieve the dataframe with psis or pval for each feature and a `Warning` if the psi exceeds thresholds or if the pval doesn't exceeds alpha.

    Args:
        base_df (pd.DataFrame/dict): reference dataset, usually the train set.
        compare_df (pd.DataFrame/dict): comparison dataset, usually the test set.
        type_data (str, optional): indicates the type of input data among "data" (pd.DataFrame), "metadata" (dict of metadata containing for each feature the bin mapping). Defaults to "data".
        feat_to_check (list, optional): list of features to be checked.
        stat (str, optional): indicates the type of test between "psi" and "pval" (Kolmogorov-Smirnov Test for numerical features and Chi-Squared Test for categorical features). Defaults to "psi".
        alpha (float, optional): level of significance for the test. Defaults to 0.05.
        max_psi (float, optional): maximum psi acceptable. Defaults to 0.2.
        mid_psi (float, optional): medium psi acceptable. Defaults to 0.1.
        psi_nbins (int, optional): number of bins into which the features will be bucketed (maximum) to compute psi. Deafults to 1000.
        psi_bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.
        return_meta_ref (bool, optional): boolean to save the reference metadata dictionary. Defaults to False.
        dim_threshold (int,optional): maximum significant size of the test set for Chi-Square test. Defaults to 5000.

    Returns:
        pd.DataFrame: report containing psis, features and warnings
        dict () : reference metadata dictionary
    """
    if feat_to_check is None:
        feat_to_check = []
    stat_results, meta_ref = retrieve_stat(
        base_df,
        compare_df,
        stat,
        feat_to_check,
        type_data=type_data,
        psi_nbins=psi_nbins,
        psi_bin_min_pct=psi_bin_min_pct,
        return_meta_ref=return_meta_ref,
        dim_threshold=dim_threshold,
    )

    report_df = pd.DataFrame.from_dict(stat_results, orient="index").reset_index()

    if stat == "psi":
        report_df.columns = [
            "feature",
            "common_psi",
            "total_psi",
            "proportion_new_data",
            "proportion_old-fashioned_data",
            "validity_warning",
            "warning_constant",
        ]
        warning = pd.Series(
            [
                "Red Alert" if i > max_psi else "Yellow Alert" if (i > mid_psi and i <= max_psi) else None
                for i in report_df["common_psi"].values
            ]
        )
    else:
        report_df.columns = [
            "feature",
            "common_pval",
            "total_pval",
            "proportion_new_data",
            "proportion_old-fashioned_data",
            "validity_warning",
            "warning_constant",
        ]
        warning = pd.Series(["Red Alert" if i < alpha else None for i in report_df["common_pval"].values])

    report_df.insert(2, "warning", warning)
    report_df.loc[~report_df["warning_constant"].isnull(), "warning"] = report_df.loc[
        ~report_df["warning_constant"].isnull(), "warning_constant"
    ]
    report_df.drop(columns="warning_constant", inplace=True)

    return report_df, meta_ref


def retrieve_stat(
    base_df,
    compare_df,
    stat,
    features,
    type_data="data",
    psi_nbins=1000,
    psi_bin_min_pct=0.04,
    return_meta_ref=False,
    dim_threshold=5000,
):
    """Retrieve stat for all the features.

    Args:
        base_df (pd.DataFrame/dict): reference dataset, usually the train set.
        compare_df (pd.DataFrame/dict): comparison dataset, usually the test set.
        stat (str): indicates the type of test between "psi" and "pval" (Kolmogorov-Smirnov Test for numerical features and Chi-Squared Test for categorical features). Default to "psi".
        features (list): list of features to be checked
        type_data (str, optional): indicates the type of input data among "data" (pd.DataFrame), "metadata" (dict of metadata containing for each feature the bin mapping). Defaults to "data".
        psi_nbins (int, optional): number of bins into which the features will be bucketed (maximum) to compute psi. Deafults to 1000.
        psi_bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.
        return_meta_ref (bool, optional): boolean to save the reference metadata dictionary. Defaults to False.
        dim_threshold (int,optional): maximum significant size of the test set for Chi-Square test. Defaults to 5000.

    Returns:
        dict: dictionary reporting for each feature (key) the psi value (value)
        dict (optional) : reference metadata dictionary
    """
    if len(features) == 0:
        if type_data == "data":
            features = base_df.columns
        else:
            features = base_df.keys()
    if isinstance(features, str):
        features = [features]

    statt = dict()
    if return_meta_ref:
        meta_ref_dict = dict()
    else:
        meta_ref_dict = None

    if type_data == "data":
        categorical_features = get_categorical_features(base_df)
    else:
        categorical_features = [x for x in base_df.keys() if base_df[x]["type"] == "categorical"]

    if stat == "psi":
        features_pb = tqdm(features, file=sys.stdout, desc="Performing psi drift", ncols=100, leave=True)
    else:
        features_pb = tqdm(features, file=sys.stdout, desc="Performing p-value", ncols=100, leave=True)

    for ix, col in enumerate(features_pb):
        validity_warning = None
        warning = None
        if return_meta_ref:
            col_dict = dict()
        if stat == "pval":
            if type_data == "data":
                n = (len(base_df[col])) - base_df[col].isnull().sum()
                m = (len(compare_df[col])) - compare_df[col].isnull().sum()
            else:
                n = base_df[col]["not_missing_values"]
                m = compare_df[col]["not_missing_values"]
        # categorical bins generation
        if col in categorical_features:
            categorical_data = True
            # original data
            if type_data == "data":
                mapper = {x: x for x in base_df[col].dropna().unique()}
                mapper = merge_categorical_bins(base_df, col, mapper, bin_min_pct=psi_bin_min_pct)
                base_bin = base_df[col].dropna().map(mapper)
                if base_bin.nunique() == 1:
                    warning = f"Information - one bucket in reference data with more than {(1-psi_bin_min_pct)*100}% of information"
                if return_meta_ref:
                    col_dict["type"] = "categorical"
                    for y in list(dict.fromkeys(mapper.values())):
                        col_dict[y] = {
                            "labels": [k for k, v in mapper.items() if v == y],
                            "freq": base_bin.value_counts(normalize=True)[y],
                        }
                if set(compare_df[col].dropna().unique()) - set(base_df[col].dropna().unique()) != set():
                    only_comp = {
                        x: "_other_"
                        for x in list(set(compare_df[col].dropna().unique()) - set(base_df[col].dropna().unique()))
                    }
                    mapper.update(only_comp)
                comp_bin = compare_df[col].dropna().map(mapper)
                if (set(base_bin.unique()) - set(comp_bin.unique()) != set()) and (
                    set(comp_bin.unique()) - set(base_bin.unique()) != set()
                ):
                    validity_warning = "Red Alert - new and old-fashioned categorical data"
                else:
                    if set(base_bin.unique()) - set(comp_bin.unique()) != set():
                        validity_warning = "Yellow Alert - old-fashioned categorical data"
                    if set(comp_bin.unique()) - set(base_bin.unique()) != set():
                        validity_warning = "Red Alert - new categorical data"
            # metadata
            else:
                base_bin = {
                    x: base_df[col][x]
                    for x in base_df[col].keys()
                    if x not in ["type", "missing_values", "not_missing_values"]
                }
                if len(base_bin.keys()) == 1:
                    warning = f"Information - one bucket in reference data with more than {(1-psi_bin_min_pct)*100}% of information"
                comp_bin = {
                    x: compare_df[col][x]
                    for x in compare_df[col].keys()
                    if x not in ["type", "missing_values", "not_missing_values"]
                }
                if (set(base_bin.keys()) - set(comp_bin.keys()) != set()) and (
                    set(comp_bin.keys()) - set(base_bin.keys()) != set()
                ):
                    validity_warning = "Red Alert - new and old-fashioned categorical data"
                else:
                    if set(base_bin.keys()) - set(comp_bin.keys()) != set():
                        validity_warning = "Yellow Alert - old-fashioned categorical data"
                    if set(comp_bin.keys()) - set(base_bin.keys()) != set():
                        validity_warning = "Red Alert - new categorical data"

        # numerical bins generation
        else:
            categorical_data = False
            # original data
            if type_data == "data":
                cuts = retrieve_bin_numerical(base_df, col, max_n_bins=psi_nbins)
                cuts = merge_numerical_bins(base_df, col, cuts, bin_min_pct=psi_bin_min_pct)
                base_bin = pd.cut(base_df[col].dropna(), cuts, right=True, precision=25, duplicates="drop")
                if len(cuts) == 2:
                    warning = f"Information - one bucket in reference data with more than {(1-psi_bin_min_pct)*100}% of information"
                comp_bin = pd.cut(
                    compare_df[col].dropna().astype("float"), cuts, right=True, precision=25, duplicates="drop"
                )
                if (compare_df[col].max() > base_df[col].max()) and (compare_df[col].min() < base_df[col].min()):
                    validity_warning = "Information - values outside the min-max range in the new data"
                else:
                    if compare_df[col].min() < base_df[col].min():
                        validity_warning = "Information - values below lower bound in the new data"
                    if compare_df[col].max() > base_df[col].max():
                        validity_warning = "Information - values above the upper bound in the new data"
                if return_meta_ref:
                    col_dict["type"] = "numerical"
                    col_dict["min_val"] = base_df[col].min()
                    col_dict["max_val"] = base_df[col].max()
                    col_dict["not_missing_values"] = len(base_df[col]) - base_df[col].isnull().sum()
                    for i, y in enumerate(base_bin.value_counts(normalize=True).sort_index().index):
                        col_dict[f"bin_{i}"] = {
                            "min": y.left,
                            "max": y.right,
                            "freq": base_bin.value_counts(normalize=True)[y],
                        }
            # metadata
            else:
                base_bin = {
                    x: base_df[col][x]
                    for x in base_df[col].keys()
                    if x not in ["type", "missing_values", "min_val", "max_val", "not_missing_values"]
                }
                if len(base_bin.keys()) == 1:
                    warning = f"Information - one bucket in reference data with more than {(1-psi_bin_min_pct)*100}% of information"
                comp_bin = {
                    x: compare_df[col][x]
                    for x in compare_df[col].keys()
                    if x not in ["type", "missing_values", "min_val", "max_val", "not_missing_values"]
                }
                if (compare_df[col]["max_val"] > base_df[col]["max_val"]) and (
                    compare_df[col]["min_val"] < base_df[col]["min_val"]
                ):
                    validity_warning = "Information - values outside the min-max range in the new data"
                else:
                    if compare_df[col]["min_val"] < base_df[col]["min_val"]:
                        validity_warning = "Information - values below lower bound in the new data"
                    if compare_df[col]["max_val"] > base_df[col]["max_val"]:
                        validity_warning = "Information - values above the upper bound in the new data"

        if return_meta_ref:
            meta_ref_dict[col] = col_dict
        if (stat == "psi") and (ix == len(features) - 1):
            features_pb.set_description("Completed psi drift", refresh=True)
        elif (stat == "pval") and (ix == len(features) - 1):
            features_pb.set_description("Completed p-value", refresh=True)

        if stat == "psi":
            # drift indicators generation for psi
            common_psi, total_psi, prop_new_data, prop_oldfash_data = stat_value(
                base_bin, comp_bin, stat=stat, type_data=type_data, categorical=categorical_data
            )
            statt[col] = [common_psi, total_psi, prop_new_data, prop_oldfash_data, validity_warning, warning]
        else:
            # drift indicators generation for pval
            common_pval, total_pval, prop_new_data, prop_oldfash_data = stat_value(
                base_bin, comp_bin, stat=stat, n=n, m=m, type_data=type_data, categorical=categorical_data
            )
            statt[col] = [common_pval, total_pval, prop_new_data, prop_oldfash_data, validity_warning, warning]

    if return_meta_ref:
        if type_data == "metadata":
            meta_ref_dict = base_df

    if (stat == "pval") and (len(compare_df) > dim_threshold):
        warnings.warn(
            "Test chi quadro per le variabili categoriche potrebbe non essere affidabile se la dimensione del test set è maggiore di 5000"
        )

    return statt, meta_ref_dict


def stat_value(ref, new, stat, n=None, m=None, type_data="data", categorical=True):
    """Retrieve the stat and proportion of new and old-fashioned data.

    Args:
        ref (pd.Series/dict): series of the bins in the base/reference feature.
        new (pd.Series/dict): series of the bins in the new/comparison feature.
        stat (str): indicates the type of test between "psi" and "pval" (Kolmogorov-Smirnov Test for numerical features and Chi-Squared Test for categorical features). Default to "psi".
        n (int, optional): numbers of non missing values in the base/reference feature. Defaults to None.
        m (int, optional): numbers of non missing values in the new/comparison feature. Defaults to None.
        type_data (str, optional): indicates the type of input data among "data" (pd.DataFrame), "metadata" (dict of metadata containing for each feature the bin mapping). Defaults to "data".
        categorical (bool, optional): indicates whether the bins come from a categorical or numerical variable. Defaults to True.

    Returns:
        float: the common stat
        float: the total stat
        float: the proportion of new data
        float: the proportion of old-fashioned data
    """
    if type_data == "data":
        ref_count = pd.DataFrame(ref.value_counts(normalize=True))
        ref_count.columns = ["ref"]
        new_count = pd.DataFrame(new.value_counts(normalize=True))
        new_count.columns = ["new"]
    else:
        if categorical:
            _cols = ["labels", "freq"]
        else:
            _cols = ["min", "max", "freq"]
        ref_count = pd.DataFrame.from_dict(ref, orient="index", columns=_cols).rename(columns={"freq": "ref"})
        new_count = pd.DataFrame.from_dict(new, orient="index", columns=_cols).rename(columns={"freq": "new"})

    # common stat
    merged_common = create_stat_table(ref_count, new_count, stat, n, m, "inner", categorical, type_data)

    # total stat
    merged_total = create_stat_table(ref_count, new_count, stat, n, m, "outer", categorical, type_data)

    # proportion old-fashioned data
    prop_oldfash_data = merged_total.loc[
        [x for x in list(ref_count.index) if x not in list(new_count.index)], "ref"
    ].sum()

    # proportion new data
    prop_new_data = merged_total.loc[[x for x in list(new_count.index) if x not in list(ref_count.index)], "new"].sum()

    if stat == "psi":
        return merged_common["psi"].sum(), merged_total["psi"].sum(), prop_new_data, prop_oldfash_data
    else:
        # Chi-Square p-value for categorical features
        if categorical:
            stat_cat_c = merged_common["pval"].sum()
            stat_cat_t = merged_total["pval"].sum()
            return (
                stats.chi2.sf(stat_cat_c, len(merged_common["pval"]) - 1),
                stats.chi2.sf(stat_cat_t, len(merged_total["pval"]) - 1),
                prop_new_data,
                prop_oldfash_data,
            )
        # KS p-value for numerical features
        else:
            stat_num_c = max(merged_common["pval"])
            stat_num_t = max(merged_total["pval"])
            # degrees of freedom for KS Test
            en = np.round((n * m) / (n + m))
            return stats.kstwo.sf(stat_num_c, en), stats.kstwo.sf(stat_num_t, en), prop_new_data, prop_oldfash_data


def dd_red_yellow(df, missing_alert=False, missing_max_thresholds=10, missing_mid_thresholds=5):
    """Color report of Class Data Drift based on Alert.

    Args:
        df (pd.DataFrame): report in input.
        missing_alert (bool, optional): boolean to color report based on alert on missing values. Defaults to False.
        missing_max_thresholds (float, optional): maximun percentage of new missing values threshold, computed if missing_alert is set to True. Defaults to 10.
        missing_mid_thresholds (float, optional): medium percentage of new missing values threshold, computed if missing_alert is set to True. Defaults to 5.

    Returns:
        pd.DataFrame: color-mapping report
    """
    ret = pd.DataFrame("", index=df.index, columns=df.columns)

    ret.loc[df.warning == "Red Alert", ["common_psi", "warning"]] = "background-color: red"
    ret.loc[df.warning == "Yellow Alert", ["common_psi", "warning"]] = "background-color: yellow"
    ret.loc[
        df["validity_warning"].fillna("Missing").str.startswith("Red Alert"), "validity_warning"
    ] = "background-color: red"
    ret.loc[
        df["validity_warning"].fillna("Missing").str.startswith("Yellow Alert"), "validity_warning"
    ] = "background-color: yellow"
    if missing_alert:
        ret.loc[df["drift_perc_missing"] > missing_max_thresholds, "drift_perc_missing"] = "background-color: red"
        ret.loc[
            df["drift_perc_missing"] <= missing_max_thresholds and df["drift_perc_missing"] > missing_mid_thresholds,
            "drift_perc_missing",
        ] = "background-color: yellow"

    return ret


def create_stat_table(
    ref_count, new_count, stat, n=None, m=None, merge_type="inner", categorical=True, type_data="data"
):
    """Create stat table using normalized frequency on bins of a feature in reference and new datasets.

    Args:
        ref_count (pd.DataFrame): dataframe counting relative frequency on bins of a feature in the reference dataset.
        new_count (pd.DataFrame): dataframe counting relative frequency on bins of a feature in the new dataset.
        stat (str): indicates the type of test between "psi" and "pval" (Kolmogorov-Smirnov Test for numerical features and Chi-Squared Test for categorical features). Default to "psi".
        n (int, optional): numbers of non missing values in the base/reference feature. Defaults to None.
        m (int, optional): numbers of non missing values in the new/comparison feature. Defaults to None.
        merge_type (str): type of merge to be performed between "left", "right", "outer", "inner" and "cross". Defaults to "inner".
        categorical (bool, optional): indicates whether the bins come from a categorical or numerical variable. Defaults to True.
        type_data (str, optional): indicates the type of input data among "data" (pd.DataFrame), "metadata" (dict of metadata containing for each feature the bin mapping). Defaults to "data".

    Returns:
        pd.DataFrame: stat table
    """
    if type_data == "data":
        merged = ref_count.merge(new_count, how=merge_type, left_index=True, right_index=True).fillna(0).sort_index()
    else:
        if not categorical:
            ref_count = ref_count.sort_values(by="min")
            merged = ref_count.merge(new_count, how=merge_type, left_index=True, right_index=True).fillna(0)
        else:
            merged = (
                ref_count.merge(new_count, how=merge_type, left_index=True, right_index=True).fillna(0).sort_index()
            )

    if stat == "psi":
        merged[["ref", "new"]] = merged[["ref", "new"]].fillna(0).clip(lower=0.000001)
        merged["psi"] = (merged["new"] - merged["ref"]) * (np.log(merged["new"] / merged["ref"]))
    else:
        # Chi-Square statistic for categorical features
        if categorical:
            merged["ref"] = np.round(merged["ref"] * n)
            merged["new"] = np.round(merged["new"] * m)
            merged["ref"] = np.round((merged["ref"] / n) * m)
            merged["pval"] = ((merged["new"] - merged["ref"]) ** 2) / (merged["ref"])
        # KS statistic for numerical features
        else:
            merged["ref"] = np.cumsum(merged["ref"])
            merged["new"] = np.cumsum(merged["new"])
            merged["pval"] = [np.abs(x - y) for x, y in zip(merged["ref"], merged["new"])]

    return merged



import pandas as pd
import numpy as np
import warnings

from model_monitoring.utils import check_features_sets

from model_monitoring.config import read_config

params = read_config(config_dir="config", name_params="params.yml")
standard_threshold = params["xai_threshold"]
standard_psi_threshold = params["data_drift_threshold"]


class XAIDrift:
    """XAI Drift Class."""

    def __init__(self, xai_curr, xai_stor, feat_to_check=None, config_threshold=None):
        """XAI Drift Class.

        Args:
            xai_curr (dict): current xai dictionary.
            xai_stor (dict): historical xai dictionary.
            feat_to_check (list, optional): list of features to be checked. Deafualts to None.
            config_threshold (dict, optional): dictionary containing relative and absolute tolerances threshold settings. Defaults to None.
        """
        # Set configuration for relative and asbolute tolerance threshold for alerting
        if config_threshold is None:
            config_threshold = standard_threshold
        self.config_theshold = config_threshold

        # Check if the scores are assigned to the same set of features
        check_features_sets(
            features_1=list(xai_curr["feat_importance"].keys()), features_2=list(xai_stor["feat_importance"].keys())
        )

        if feat_to_check is not None:
            self.xai_curr = {
                "type": xai_curr["type"],
                "feat_importance": {x: xai_curr["feat_importance"][x] for x in feat_to_check},
            }
            self.xai_stor = {
                "type": xai_stor["type"],
                "feat_importance": {x: xai_stor["feat_importance"][x] for x in feat_to_check},
            }
        else:
            self.xai_curr = xai_curr
            self.xai_stor = xai_stor

        self.feat_to_check = feat_to_check

        # Check if the type of feature importance of historical and current xai model is the same
        if self.xai_curr["type"] != self.xai_stor["type"]:
            if (self.xai_curr["type"] == "coef") or (self.xai_stor["type"] == "coef"):
                raise ValueError(
                    f"'{self.xai_curr['type']}' type of feature importance in current xai model is not compatible with '{xai_stor['type']}' typ in historical xai model"
                )
            else:
                warnings.warn(
                    "the type of feature importance in current and historical xai model are not the same but they are compatible"
                )

        # Initialize the report
        xai_drift_report = (
            pd.DataFrame.from_dict(self.xai_curr["feat_importance"], "index", columns=["curr_score"])
            .reset_index()
            .rename(columns={"index": "feature"})
        )
        xai_stor_report = (
            pd.DataFrame.from_dict(self.xai_stor["feat_importance"], "index", columns=["stor_score"])
            .reset_index()
            .rename(columns={"index": "feature"})
        )
        # Save features not in common in both xai model dictionaries
        self.feat_only_cur = list(set(xai_drift_report.feature.unique()) - set(xai_stor_report.feature.unique()))
        self.feat_only_stor = list(set(xai_stor_report.feature.unique()) - set(xai_drift_report.feature.unique()))

        xai_drift_report = xai_drift_report.merge(xai_stor_report, how="outer", on="feature").fillna(0)

        self.xai_drift_report = xai_drift_report
        self.xai_psi = False

    def get_drift(self, xai_psi=True, psi_config_threshold=None):
        """Load on the report the feature importance drift and relative alert from current and historical XAI model.

        Args:
            xai_psi (bool, optional): retrieve psi for feature importances. Defaults to True.
            psi_config_threshold (dict, optional): dictionary containing psi threshold settings. Defaults to None.
        """
        self.relative_red = self.config_theshold["relative_red"]
        self.relative_yellow = self.config_theshold["relative_yellow"]
        self.absolute_tol = self.config_theshold["absolute_tol"]

        # Verify if 'control_type' column exist in report
        if "control_type" not in self.xai_drift_report.columns:
            # if control_type column doesn't exist, it is created in first position
            self.xai_drift_report.insert(0, "control_type", "feature drift perc")

        self.xai_psi = xai_psi
        if self.xai_psi:
            # Set thresholds for PSI
            if psi_config_threshold is None:
                psi_config_threshold = standard_psi_threshold
            self.psi_config_threshold = psi_config_threshold
            # Add 'xai_psi' feature if it doesn't exist yet
            if "xai_psi" not in self.xai_drift_report.feature.values:
                self.xai_drift_report = pd.concat(
                    [self.xai_drift_report, pd.DataFrame({"feature": ["xai_psi"]})], ignore_index=True
                )
                self.xai_drift_report.loc[lambda x: x.feature == "xai_psi", "control_type"] = "psi xai score"
        else:
            if "xai_psi" in self.xai_drift_report.feature.values:
                self.xai_drift_report = self.xai_drift_report.loc[lambda x: x.feature != "xai_psi"]

        # Generation Drift
        for a in self.xai_drift_report.feature.values:
            stor_score = self.xai_drift_report.loc[self.xai_drift_report.feature == a, "stor_score"].values[0]
            curr_score = self.xai_drift_report.loc[self.xai_drift_report.feature == a, "curr_score"].values[0]
            self.xai_drift_report.loc[self.xai_drift_report.feature == a, "value"] = (
                (curr_score - stor_score) / stor_score * 100 if stor_score != 0 else 0
            )
            if a == "xai_psi":
                psi_score = (
                    self.xai_drift_report.loc[lambda x: x.feature != "xai_psi", ["curr_score", "stor_score"]]
                    .clip(lower=0.000001)
                    .assign(psi=lambda x: (x.curr_score - x.stor_score) * (np.log(x.curr_score / x.stor_score)))
                    .loc[:, "psi"]
                    .sum()
                )
                self.xai_drift_report.loc[self.xai_drift_report.feature == a, "value"] = psi_score

        # Generation Alert
        for a in self.xai_drift_report.feature.values:
            stor_score = self.xai_drift_report.loc[self.xai_drift_report.feature == a, "stor_score"].values[0]
            curr_score = self.xai_drift_report.loc[self.xai_drift_report.feature == a, "curr_score"].values[0]
            if a in self.feat_only_cur:
                self.xai_drift_report.loc[
                    self.xai_drift_report.feature == a, "relative_warning"
                ] = "Feature not used in historical model"
            elif a in self.feat_only_stor:
                self.xai_drift_report.loc[
                    self.xai_drift_report.feature == a, "relative_warning"
                ] = "Feature not used in current model"
            else:
                if abs(curr_score - stor_score) >= self.absolute_tol:
                    drift_xai = self.xai_drift_report.loc[self.xai_drift_report.feature == a, "value"].values[0]
                    if abs(drift_xai) > self.relative_red * 100:
                        self.xai_drift_report.loc[self.xai_drift_report.feature == a, "relative_warning"] = "Red Alert"
                    else:
                        if (abs(drift_xai) < self.relative_red * 100) and (abs(drift_xai) > self.relative_yellow * 100):
                            self.xai_drift_report.loc[
                                self.xai_drift_report.feature == a, "relative_warning"
                            ] = "Yellow Alert"
                        else:
                            self.xai_drift_report.loc[self.xai_drift_report.feature == a, "relative_warning"] = np.nan
                else:
                    self.xai_drift_report.loc[self.xai_drift_report.feature == a, "relative_warning"] = np.nan
            if a == "xai_psi":
                if psi_score > self.psi_config_threshold["max_psi"]:
                    self.xai_drift_report.loc[self.xai_drift_report.feature == a, "relative_warning"] = "Red Alert"
                else:
                    if (psi_score <= self.psi_config_threshold["max_psi"]) and (
                        psi_score > self.psi_config_threshold["mid_psi"]
                    ):
                        self.xai_drift_report.loc[
                            self.xai_drift_report.feature == a, "relative_warning"
                        ] = "Yellow Alert"
                    else:
                        self.xai_drift_report.loc[self.xai_drift_report.feature == a, "relative_warning"] = np.nan

    def get_report(self):
        """Return the xai drift report.

        Returns:
            pd.DataFrame: report of the class.
        """
        return self.xai_drift_report.sort_values("stor_score", key=abs)

    def plot(self):
        """Plot the report on features importance drift."""
        self.xai_drift_report.loc[lambda x: x.feature != "xai_psi"].sort_values("stor_score", key=abs).plot(
            x="feature", y=["curr_score", "stor_score"], kind="barh", title="Features Importance Drift"
        )



def compute_metric(y_true, y_pred, metric, **kwargs):
    """Compute the supervised metric scoring.

    Args:
        y_true (pd.Series/np.array): true values of target
        y_pred (pd.Series/np.array): predictions/scores of target
        metric (function): metric function

    Returns:
        metric scoring performance
    """
    return metric(y_true, y_pred, **kwargs)



import pandas as pd
from typing import Dict

from model_monitoring.data_drift.data_drift import stat_report
from model_monitoring.utils import check_features_sets, convert_Int_dataframe, convert_Int_series

from model_monitoring.config import read_config

params = read_config(config_dir="config", name_params="params.yml")
standard_threshold_psi = params["data_drift_threshold"]


class DataDrift:
    """Data Drift Class."""

    def __init__(self, data_stor, data_curr, type_data="auto", feat_to_check=None, config_threshold=None):
        """Data Drift Class.

        Args:
            data_stor (pd.DataFrame/dict): historical dataset.
            data_curr (pd.DataFrame/dict): current dataset.
            type_data (str, optional): indicates the type of input data among "data" (pd.DataFrame), "metadata" (dict of metadata containing for each feature the bin mapping) and "auto".
                If "data" is provided as input the class computes data drift on orginal data, while for "metadata" input it computes data drift on information contained in dictionaries.
                Defaults to "auto".
            feat_to_check (list, optional): list of features to be checked. Deafualts to None.
            config_threshold (dict, optional): dictionary containing psi threshold settings. Defaults to None.
        """
        # Set psi threshold
        if config_threshold is None:
            config_threshold = standard_threshold_psi
        self.config_threshold = config_threshold

        if type_data not in ["auto", "data", "metadata"]:
            raise ValueError(
                f"{type_data} is not a valid type_data. It should be one of the following:\n ['auto','data','metadata']"
            )

        if type_data == "auto":
            if isinstance(data_stor, pd.DataFrame) and isinstance(data_curr, pd.DataFrame):
                type_data = "data"
            elif isinstance(data_stor, Dict) and isinstance(data_curr, Dict):
                type_data = "metadata"
            else:
                raise ValueError("format data inputs are not valid. Choose both pd.DataFrame or both dict")
        self.type_data = type_data

        if type_data == "data":
            check_features_sets(features_1=list(data_stor.columns), features_2=list(data_curr.columns))
            list_com_features = list(set(data_stor.columns).intersection(set(data_curr.columns)))
        else:
            check_features_sets(features_1=list(data_stor.keys()), features_2=list(data_curr.keys()))
            list_com_features = list(set(data_stor.keys()).intersection(set(data_curr.keys())))
        if len(list_com_features) == 0:
            raise ValueError("No features in common between the two sets")

        if feat_to_check is None:
            feat_to_check = list_com_features

        self.feat_to_check = feat_to_check

        if isinstance(self.feat_to_check, str):
            self.feat_to_check = [self.feat_to_check]

        self.data_stor_original = data_stor
        self.data_curr_original = data_curr
        if type_data == "data":
            if len(self.feat_to_check) == 1:
                self.data_stor = pd.DataFrame(convert_Int_series(data_stor[self.feat_to_check[0]]))
                self.data_curr = pd.DataFrame(convert_Int_series(data_curr[self.feat_to_check[0]]))
            else:
                self.data_stor = convert_Int_dataframe(data_stor[self.feat_to_check])
                self.data_curr = convert_Int_dataframe(data_curr[self.feat_to_check])
        else:
            self.data_stor = {x: data_stor[x] for x in self.feat_to_check}
            self.data_curr = {x: data_curr[x] for x in self.feat_to_check}

        # Check valid format for metadata
        if type_data == "metadata":
            for x in self.feat_to_check:
                if "type" not in self.data_stor[x].keys():
                    raise ValueError(f"not type provided for {x} feature in reference metadata")
                elif "type" not in self.data_curr[x].keys():
                    raise ValueError(f"not type provided for {x} feature in new metadata")
                else:
                    if self.data_stor[x]["type"] not in ["categorical", "numerical"]:
                        raise ValueError(
                            f"{self.data_stor[x]['type']} not valid type for {x} feature. Choose among ['categorical','numerical']"
                        )
                    if self.data_curr[x]["type"] not in ["categorical", "numerical"]:
                        raise ValueError(
                            f"{self.data_curr[x]['type']} not valid type for {x} feature. Choose among ['categorical','numerical']"
                        )
                    if self.data_stor[x]["type"] == "numerical":
                        if "min_val" not in self.data_stor[x].keys():
                            raise ValueError(f"not min_val provided for {x} numerical feature in reference metadata")
                        if "max_val" not in self.data_stor[x].keys():
                            raise ValueError(f"not max_val provided for {x} numerical feature in reference metadata")
                        for y in [
                            k
                            for k in self.data_stor[x].keys()
                            if k not in ["type", "min_val", "max_val", "missing_values", "not_missing_values"]
                        ]:
                            if "min" not in self.data_stor[x][y].keys():
                                raise ValueError(
                                    f"not min provided for {x} numerical feature for {y} bin in reference metadata"
                                )
                            if "max" not in self.data_stor[x][y].keys():
                                raise ValueError(
                                    f"not max provided for {x} numerical feature for {y} bin in reference metadata"
                                )
                            if "freq" not in self.data_stor[x][y].keys():
                                raise ValueError(
                                    f"not freq provided for {x} numerical feature for {y} bin in reference metadata"
                                )
                    if self.data_curr[x]["type"] == "numerical":
                        if "min_val" not in self.data_curr[x].keys():
                            raise ValueError(f"not min_val provided for {x} numerical feature in new metadata")
                        if "max_val" not in self.data_curr[x].keys():
                            raise ValueError(f"not max_val provided for {x} numerical feature in new metadata")
                        for y in [
                            k
                            for k in self.data_curr[x].keys()
                            if k not in ["type", "min_val", "max_val", "missing_values", "not_missing_values"]
                        ]:
                            if "min" not in self.data_curr[x][y].keys():
                                raise ValueError(
                                    f"not min provided for {x} numerical feature for {y} bin in new metadata"
                                )
                            if "max" not in self.data_curr[x][y].keys():
                                raise ValueError(
                                    f"not max provided for {x} numerical feature for {y} bin in new metadata"
                                )
                            if "freq" not in self.data_curr[x][y].keys():
                                raise ValueError(
                                    f"not freq provided for {x} numerical feature for {y} bin in new metadata"
                                )
                    if self.data_stor[x]["type"] == "categorical":
                        for y in [
                            k
                            for k in self.data_stor[x].keys()
                            if k not in ["type", "missing_values", "not_missing_values"]
                        ]:
                            if "labels" not in self.data_stor[x][y].keys():
                                raise ValueError(
                                    f"not labels provided for {x} categorical feature for {y} bin in reference metadata"
                                )
                            if "freq" not in self.data_stor[x][y].keys():
                                raise ValueError(
                                    f"not freq provided for {x} categorical feature for {y} bin in reference metadata"
                                )
                    if self.data_curr[x]["type"] == "categorical":
                        for y in [
                            k
                            for k in self.data_curr[x].keys()
                            if k not in ["type", "missing_values", "not_missing_values"]
                        ]:
                            if "labels" not in self.data_curr[x][y].keys():
                                raise ValueError(
                                    f"not labels provided for {x} categorical feature for {y} bin in new metadata"
                                )
                            if "freq" not in self.data_curr[x][y].keys():
                                raise ValueError(
                                    f"not freq provided for {x} categorical feature for {y} bin in new metadata"
                                )

        # Initialize the report
        self.data_drift_report = None
        self.meta_ref_dict = None

    def report_drift(
        self,
        psi_nbins=1000,
        psi_bin_min_pct=0.04,
        stat="psi",
        drift_missing=True,
        return_meta_ref=False,
        dim_threshold=5000,
    ):
        """Retrieve the report with psis or pval for each feature in both dataframe sets and `Warning` alerts if the psis exceed thresholds or if the pval doesn't exceeds alpha.

        Args:
            psi_nbins (int, optional): number of bins into which the features will be bucketed (maximum) to compute psi. Defaults to 1000.
            psi_bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.
            stat (str, optional): indicates the type of stat between "psi" and "pval" (Kolmogorov-Smirnov Test for numerical features and Chi-Squared Test for categorical features). Defaults to "psi".
            drift_missing (bool, optional): boolean to add to the report information on missing values drift. Defaults to True.
            return_meta_ref (bool, optional): boolean to save as attribute of the class the reference metadata dictionary. Defaults to False.
            dim_threshold (int,optional): maximum significant size of the test set for Chi-Square test. Defaults to 5000.

        Returns:
            pd.DataFrame: report of the class.
        """
        self.psi_nbins = psi_nbins
        self.psi_bin_min_pct = psi_bin_min_pct
        if stat not in ["psi", "pval"]:
            raise ValueError(f"{stat} is not a valid test. It should be one of the following:\n ['psi','pval']")
        self.stat = stat
        self.drift_missing = drift_missing
        self.return_meta_ref = return_meta_ref
        self.dim_threshold = dim_threshold

        if self.stat == "psi":
            # data drift psi report
            data_drift_report, meta_ref_dict = stat_report(
                base_df=self.data_stor,
                compare_df=self.data_curr,
                type_data=self.type_data,
                feat_to_check=self.feat_to_check,
                stat=self.stat,
                alpha=None,
                max_psi=self.config_threshold["max_psi"],
                mid_psi=self.config_threshold["mid_psi"],
                psi_nbins=self.psi_nbins,
                psi_bin_min_pct=self.psi_bin_min_pct,
                return_meta_ref=self.return_meta_ref,
                dim_threshold=self.dim_threshold,
            )

        else:
            # data drift p-value report
            data_drift_report, meta_ref_dict = stat_report(
                base_df=self.data_stor,
                compare_df=self.data_curr,
                type_data=self.type_data,
                feat_to_check=self.feat_to_check,
                stat=self.stat,
                alpha=self.config_threshold["alpha"],
                max_psi=None,
                mid_psi=None,
                psi_nbins=self.psi_nbins,
                psi_bin_min_pct=self.psi_bin_min_pct,
                return_meta_ref=self.return_meta_ref,
                dim_threshold=self.dim_threshold,
            )

        self.meta_ref_dict = meta_ref_dict

        # missing values drift
        if drift_missing:
            if self.type_data == "data":
                perc_drift_missing = pd.Series(
                    (
                        self.data_curr[self.feat_to_check].isnull().mean()
                        - self.data_stor[self.feat_to_check].isnull().mean()
                    )
                    * 100,
                    name="drift_perc_missing",
                )
                if self.return_meta_ref:
                    for col in self.meta_ref_dict:
                        self.meta_ref_dict[col]["missing_values"] = self.data_stor[col].isnull().mean()
            else:
                missing_dict = dict()
                for col in self.feat_to_check:
                    try:
                        missing_dict[col] = (
                            self.data_curr[col]["missing_values"] - self.data_stor[col]["missing_values"]
                        ) * 100
                    except Exception:
                        raise ValueError(f"no missing values provided for {col} feature")
                perc_drift_missing = pd.DataFrame.from_dict(
                    missing_dict, orient="index", columns=["drift_perc_missing"]
                )
            data_drift_report = data_drift_report.merge(perc_drift_missing, left_on="feature", right_index=True)
        self.data_drift_report = data_drift_report
        return self.data_drift_report

    def get_meta_ref(self):
        """Return the reference metadata dictionary.

        Returns:
            dict: reference metadata dictionary.
        """
        return self.meta_ref_dict



import warnings
import re
import pandas as pd
import numpy as np
import shap
from functools import reduce

from sklearn.metrics import r2_score, roc_auc_score, balanced_accuracy_score
from sklearn.model_selection import RandomizedSearchCV, StratifiedKFold, KFold, GroupKFold
from sklearn.preprocessing import OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.inspection import permutation_importance
from lightgbm import LGBMClassifier, LGBMRegressor
from model_monitoring.utils import (
    check_size,
    get_categorical_features,
    convert_Int_series,
    convert_Int_dataframe,
    absmax,
)

from model_monitoring.config import read_config

params = read_config(config_dir="config", name_params="params.yml")
default_grid = params["grid_xai"]


class XAI:
    """XAI (Explainable artificial intelligence) Class."""

    def __init__(
        self,
        model_input="default",
        model=None,
        use_cv=False,
        grid=None,
        cv_funct=RandomizedSearchCV,
        cv_scoring="auto",
        n_iter=40,
        cv_type=StratifiedKFold(5, random_state=42, shuffle=True),
        algo_type="auto",
        dim_cat_threshold=None,
    ):
        """XAI (Explainable artificial intelligence) Class.

        Args:
            model_input (str, optional): "default", "custom", determines if using a Gradient Boosting model with "default"
                or a model given in input with "custom". Defaults to "default".
            model : classifier or regressor in sklearn API class. Defaults to None.
            use_cv (bool, optional): determines if using hyperparameters tuning with CV logic. Defaults to False.
            grid (dict, optional): hyperparameters grid. Defaults to None.
            cv_funct: function or class for the Cross Validation. Defaults to None.
            cv_scoring: scoring argument of the cv_funct. Defaults to "auto" selects "roc_auc" fo classification, "r2" for regression
                and "balanced accuracy" for multiclass.
            n_iter (int, optional): number of iteration, i.e. set of hyperparams tested in Cross Validation. Defaults to 20.
            cv_type : function or class for defining the CV sets. Defaults to StratifiedKFold(5, random_state=42, shuffle=True).
            algo_type (str, optional): "auto", "classification", "multiclass", "regression", describes the problem type.
                "classification" has to be used only for binary classification. Defaults to "auto".
            dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not Top-OHE.
                If None standard OHE is imputed for categorical features. Defaults to None.
        """
        if model_input not in ["default", "custom"]:
            raise ValueError("model_input argument must be one of ['default','custom']")

        # Attributes
        self.model_input = model_input
        self.model = model
        self.use_cv = use_cv
        self.grid = grid
        self.cv_funct = cv_funct
        self.cv_scoring = cv_scoring
        self.n_iter = n_iter
        self.cv_type = cv_type
        self.algo_type = algo_type
        self.dim_cat_threshold = dim_cat_threshold
        self._fitted = False
        self.report_feat_imp = None
        self.dict_report_feat_imp = None

    def fit(self, db, output_model, manage_groups=False, groups=None):
        """Fit data given in input with the model provided in input in the class.

        Args:
            db (pd.Series/pd.DataFrame): Features used from the model to explain.
            output_model (pd.Series/np.array): output of the model to explain.
            manage_groups (bool, optional): determines if there is a feature whose groups have to be kept joined in CV. Defaults to False.
            groups (pd.Series, optional): feature whose groups have to be kept joined in CV. Defaults to None.
        """
        # Check size of the output of the model and dataset in input
        check_size(db, output_model)

        self.db = convert_Int_dataframe(db)
        self.output_model = convert_Int_series(output_model)

        # Set algo_type in 'auto' mode
        vals = len(np.unique(self.output_model))
        if vals == 1:
            warnings.warn("The output model selected is constant")
        else:
            if self.algo_type == "auto":
                if vals <= 2:
                    self.algo_type = "classification"
                elif vals < 11:
                    self.algo_type = "multiclass"
                else:
                    self.algo_type = "regression"
            elif self.algo_type not in ["classification", "regression", "multiclass"]:
                raise ValueError(
                    "algo_type argument must be one of ['auto', 'classification', 'regression', 'multiclass']"
                )

        # Set model in 'default' mode
        if self.model_input == "default":
            if self.algo_type == "regression":
                self.model = LGBMRegressor()
            else:
                self.model = LGBMClassifier()

        # If using a CV strategy, define the model as the SearchCV, otherwise initialize the model provided in input as attribute
        if self.use_cv:
            if self.model_input == "default":
                if self.algo_type == "regression":
                    self.grid = default_grid[self.algo_type]["grid_model"]
                else:
                    self.grid = default_grid["classification"]["grid_model"]
            if self.grid is None:
                self.grid = {}
            if self.cv_scoring == "auto":
                if self.algo_type == "classification":
                    self.cv_scoring = "roc_auc"
                elif self.algo_type == "regression":
                    self.cv_scoring = "r2"
                else:
                    self.cv_scoring = "balanced_accuracy"
            if (self.algo_type == "regression") and (
                str(self.cv_type.__class__()).startswith(("StratifiedKFold", "StratifiedGroupKFold"))
            ):
                self.cv_type = KFold(
                    n_splits=self.cv_type.n_splits, random_state=self.cv_type.random_state, shuffle=self.cv_type.shuffle
                )
                warnings.warn("Fold Cross Validation uncorrect for regression algorithm, KFold is set")
            try:
                CVSel_algo = self.cv_funct(
                    self.model, self.grid, n_iter=self.n_iter, cv=self.cv_type, scoring=self.cv_scoring
                )
            except Exception:
                CVSel_algo = self.cv_funct(
                    self.model, self.grid, cv=self.cv_type, scoring=self.cv_scoring
                )  # for GridSearchCV
            self.model = CVSel_algo

        # Groups check
        if manage_groups:
            if groups is None:
                warnings.warn("no group defined")
                manage_groups = False
            else:
                if not groups.index.equals(self.db.index):
                    raise ValueError("Groups Series index do not match with DataFrame index in input!")
        else:
            groups = None

        self.manage_groups = manage_groups
        if groups is not None:
            self.groups = convert_Int_series(groups)
        else:
            self.groups = groups

        self.categorical_ohe = False
        db_tsf = self.db.copy()

        if self.model_input == "default":
            # Convert object type columns in category type columns for LGBM
            db_tsf[db_tsf.select_dtypes(["object"]).columns] = db_tsf.select_dtypes(["object"]).apply(
                lambda x: x.astype("category")
            )
        else:
            cats_feat = get_categorical_features(self.db)
            self.cats_feat = cats_feat

            # Fillna with Missing and 0
            db_tsf[db_tsf.select_dtypes(["category"]).columns] = db_tsf.select_dtypes(["category"]).apply(
                lambda x: x.astype("object")
            )
            db_tsf[cats_feat] = db_tsf[cats_feat].fillna("Missing")
            db_tsf[[x for x in list(set(db_tsf.columns) - set(cats_feat))]] = db_tsf[
                [x for x in list(set(db_tsf.columns) - set(cats_feat))]
            ].fillna(0)

            # One-Hot-Encoding Categorical features
            if len(cats_feat) > 0:
                cats_ohe_feat = list(self.cats_feat)
                if self.dim_cat_threshold is not None:  # for Top-OHE
                    for col in cats_ohe_feat:
                        if db_tsf[col].nunique() > self.dim_cat_threshold:
                            values_cat = list(
                                db_tsf.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                            )[: self.dim_cat_threshold]
                            for val in values_cat:
                                db_tsf[col + "_" + str(val)] = (db_tsf[col] == val).astype("int")
                            db_tsf = db_tsf.drop(columns=col)
                            cats_ohe_feat.remove(col)
                preprocessor = ColumnTransformer(
                    transformers=[
                        ("cat", OneHotEncoder(handle_unknown="ignore", sparse=False), cats_ohe_feat),
                    ],
                    remainder="passthrough",
                )
                preprocessor.fit(db_tsf.fillna(0))
                feat_names = preprocessor.get_feature_names_out()
                feat_names = [re.sub(r"((remainder)|(cat))__", "", x) for x in feat_names]
                db_tsf = pd.DataFrame(preprocessor.transform(db_tsf), columns=feat_names)
                self.categorical_ohe = True

        self.db_tsf = db_tsf

        if self.use_cv:
            # Redifine KFold strategy if there are groups to consider
            if self.manage_groups:
                if len(self.groups) != len(self.db_tsf):
                    raise ValueError(
                        "dataset to be performed shap and groups-Series don't have the same number of rows ({0},{1})".format(
                            len(self.db_tsf), len(self.groups)
                        )
                    )
                number_splits = self.cv_type.n_splits
                if not (str(self.cv_type.__class__()).startswith(("GroupKFold", "StratifiedGroupKFold"))):
                    warnings.warn("GroupKFold is set for K Fold Cross Validation strategy with managing groups")
                    self.model.cv = GroupKFold(number_splits).split(self.db_tsf, self.output_model, self.groups)

        # Fit
        self.model_fitted = self.model.fit(self.db_tsf, self.output_model)
        if self.use_cv:
            self.model_fitted = self.model_fitted.best_estimator_
        self._fitted = True

        # Initialize the xai report dictionary
        self.dict_report_feat_imp = {}

    def get_feature_importance(
        self,
        feat_imp_mode="shap",
        shap_type="tree",
        n_weighted_kmeans=10,
        n_samples_deep=1000,
        n_repeats_permutation=5,
        perm_crit="mean",
    ):
        """Retrieve the features importance dictionary.

        Args:
            feat_imp_mode (str, optional): "coef","shap","permutation","gini" describes the type of retrieving features importance:
                - "coef" retrieves the coefficient for linear models (e.g. LogisticRegression, LinearRegression, SVM with linear kernel);
                - "shap" retrieves the features' Shapley Value;
                - "permutation" retrieves features importance by permutation algorithm;
                - "gini" retrieves features importance by impurity decrease algorithm.
                Defaults to "shap".
            shap_type (str, optional): "tree","kernel","deep" describes the type of Explainer for features' Shapley Value:
                - "tree" to explain the output of ensemble tree models;
                - "kernel" to explain the output of any models using a special weighted linear regression;
                - "deep" to approximate the output of deep learning models.
                It is set only if feat_impo_mode is "shap". Defaults to 'tree'.
            n_weighted_kmeans (int, optional): number of weighted centroids for summarizing the dataset used as background dataset for integrating out features.
                It is set only if feat_impo_mode is "shap" and shap_type is "kernel". Defaults to 10.
            n_samples_deep (int, optional): number of random background samples used for integrating out features.
                It is set only if feat_impo_mode is "shap" and shap_type is "deep". Defaults to 1000.
            n_repeats_permutation (int, optional): number of times to permute a feature.
                It is set only if feat_impo_mode is "permutation". Defaults to 5.
            perm_crit (str, optional): "mean","max","min" describes the mode of aggregating the permutation importances of samples for each feature. Defaults to 'mean'.

        Returns:
            dict: features importance dictionary
        """
        # Check the type fo retrieving features importance
        if feat_imp_mode not in ["coef", "shap", "permutation", "gini"]:
            raise ValueError("feat_imp_mode argument must be one of ['coef', 'shap', 'permutation','gini']")
        self.feat_imp_mode = feat_imp_mode

        if self._fitted:
            feature_names = self.db_tsf.columns
            # Gini algorithm
            if self.feat_imp_mode == "gini":
                try:
                    feature_importances = np.abs(self.model_fitted.feature_importances_)
                except Exception:
                    raise ValueError(
                        f"{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                    )
            # Permutation algorithm
            elif self.feat_imp_mode == "permutation":
                if perm_crit not in ["mean", "max", "min"]:
                    raise ValueError("perm_crit argument must be one of ['mean','max','min']")
                self.perm_crit = perm_crit
                try:
                    self.n_repeats_permutation = n_repeats_permutation
                    perm_importance = permutation_importance(
                        self.model_fitted, self.db_tsf, self.output_model, n_repeats=n_repeats_permutation
                    )
                    # Aggregating permutation importance scores by mean
                    if self.perm_crit == "mean":
                        feature_importances = np.abs(perm_importance.importances).mean(axis=1)
                    # Aggregating permutation importance scores by max
                    elif self.perm_crit == "max":
                        feature_importances = np.abs(perm_importance.importances).max(axis=1)
                    # Aggregating permutation importance scores by min
                    elif self.perm_crit == "min":
                        feature_importances = np.abs(perm_importance.importances).min(axis=1)
                except Exception:
                    raise ValueError(
                        f"{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                    )
            # For Linear Models, retrieves absolute values of coefficients
            elif self.feat_imp_mode == "coef":
                try:
                    if self.algo_type == "regression":
                        feature_importances = self.model_fitted.coef_
                    elif self.algo_type == "classification":
                        feature_importances = self.model_fitted.coef_[0]
                    else:
                        # Scale feature importances according to frequency of the class and get the max coef value in absolute value
                        if isinstance(self.output_model, pd.Series):
                            feature_importances = absmax(
                                self.model_fitted.coef_
                                * np.reshape(
                                    self.output_model.value_counts(sort=False, normalize=True).values, (-1, 1)
                                ),
                                axis=0,
                            )
                        elif isinstance(self.output_model, np.ndarray):
                            feature_importances = absmax(
                                self.model_fitted.coef_
                                * np.reshape(
                                    pd.Series(self.output_model).value_counts(sort=False, normalize=True).values,
                                    (-1, 1),
                                ),
                                axis=0,
                            )
                        else:
                            raise ValueError(
                                f"{type(self.output_model)} is not valid type. output_model must be a pd.Series or a np.ndarray"
                            )
                except Exception:
                    raise ValueError(
                        f"{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                    )
            # Features' Shapley Values algorithm
            elif self.feat_imp_mode == "shap":
                if shap_type not in ["tree", "kernel", "deep"]:
                    raise ValueError("shap_type argument must be one of ['tree','kernel','deep']")
                self.shap_type = shap_type
                # For tree models
                if self.shap_type == "tree":
                    try:
                        feature_importances = shap.TreeExplainer(self.model_fitted).shap_values(self.db_tsf)
                    except Exception:
                        raise ValueError(
                            f"{self.shap_type}-{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                        )
                # For any models. ATTENTION: very slow, it depends on size of background dataset (second parameter of KernelExplainer)
                if self.shap_type == "kernel":
                    self.n_weighted_kmeans = n_weighted_kmeans
                    try:
                        feature_importances = shap.KernelExplainer(
                            self.model_fitted.predict, shap.kmeans(self.db_tsf, self.n_weighted_kmeans)
                        ).shap_values(self.db_tsf)
                    except Exception:
                        raise ValueError(
                            f"{self.shap_type}-{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                        )
                # For deep learning models. ATTENTION: speed depends on size of background dataset (second parameter of DeepExplainer)
                if self.shap_type == "deep":
                    self.n_samples_deep = n_samples_deep
                    try:
                        feature_importances = shap.DeepExplainer(
                            self.model_fitted, self.db_tsf.sample(n=self.n_samples_deep)
                        ).shap_values(self.db_tsf.values)[0]
                    # DeepExplainer is not compatible for tensorflow models with version upper than 2.4.0
                    except AttributeError:
                        raise ValueError(
                            "model type not currently supported! If used a tensorflow model try using the following code in importing packages phase\n\nimport tensorflow as tf\ntf.compat.v1.disable_v2_behavior()"
                        )
                    except Exception:
                        raise ValueError(
                            f"{self.shap_type}-{self.feat_imp_mode} not valid logic for retrieve feature importances with the model of the class"
                        )
                # For (multi) classification
                if isinstance(feature_importances, list):
                    db_list = list()
                    for i in range(len(feature_importances)):
                        db_list.append(
                            pd.DataFrame(
                                {
                                    "feature": feature_names,
                                    "shap_importance_" + str(i): np.abs(feature_importances[i]).mean(axis=0),
                                }
                            )
                        )
                    shap_importance = (
                        reduce(lambda left, right: pd.merge(left, right, how="outer", on="feature"), db_list)
                        .set_index("feature")
                        .assign(shap_importance=lambda x: x.sum(axis=1))
                        .loc[:, "shap_importance"]
                        .reset_index()
                    )
                    feature_importances = shap_importance.shap_importance.values
                # For regression and (some - depending on the classifier) binary classification
                elif isinstance(feature_importances, np.ndarray):
                    feature_importances = np.abs(feature_importances).mean(axis=0)
            # initialize the report of feature importances
            report_feat_imp = pd.DataFrame({"feature": feature_names, "feat_importance": feature_importances})
            # sum feature importance scores for variables auto-encoded
            if self.categorical_ohe:
                for i in self.cats_feat:
                    report_feat_imp.loc[report_feat_imp.feature.str.startswith(i)] = report_feat_imp.loc[
                        report_feat_imp.feature.str.startswith(i)
                    ].assign(
                        feature=i,
                        feat_importance=report_feat_imp.loc[
                            report_feat_imp.feature.str.startswith(i), "feat_importance"
                        ].sum(),
                    )
                report_feat_imp = report_feat_imp.drop_duplicates().dropna().reset_index(drop=True)

            # normalize with sum to 1 feature importance scores and order by feature importance
            if self.feat_imp_mode == "coef":
                self.report_feat_imp = report_feat_imp.sort_values(by=["feat_importance"], ascending=True, key=abs)
            else:
                report_feat_imp.feat_importance = (
                    report_feat_imp.feat_importance / report_feat_imp.feat_importance.sum()
                )

                self.report_feat_imp = report_feat_imp.sort_values(by=["feat_importance"], ascending=True)
        else:
            raise ValueError("no model fitted yet. Run '.fit()' with appropriate arguments before using this method.")

        self.dict_report_feat_imp["type"] = self.feat_imp_mode
        self.dict_report_feat_imp["feat_importance"] = self.report_feat_imp.set_index("feature")[
            "feat_importance"
        ].to_dict()

        return self.dict_report_feat_imp

    def get_report(self):
        """Retrieve the features importance report.

        Returns:
            pd.DataFrame: features importance dataframe
        """
        return self.report_feat_imp

    def get_score(self):
        """Get the score of XAI model, r2 for regression and roc auc score for classification ML problem.

        Returns:
            float: score of xai model
        """
        if self._fitted:
            if self.algo_type == "regression":
                model_score = r2_score(self.output_model, self.model_fitted.predict(self.db_tsf))
            elif self.algo_type == "classification":
                try:
                    model_score = roc_auc_score(self.output_model, self.model_fitted.predict_proba(self.db_tsf)[:, 1])
                except Exception:
                    model_score = balanced_accuracy_score(self.output_model, self.model_fitted.predict(self.db_tsf))
            else:
                try:
                    model_score = roc_auc_score(
                        self.output_model, self.model_fitted.predict_proba(self.db_tsf), multi_class="ovr"
                    )
                except Exception:
                    model_score = balanced_accuracy_score(self.output_model, self.model_fitted.predict(self.db_tsf))
            self.model_score = model_score
            self.dict_report_feat_imp["model_score"] = self.model_score
        else:
            raise ValueError("no model fitted yet. Run '.fit()' with appropriate arguments before using this method.")
        return self.model_score

    def plot(self):
        """Plot the report on features importance."""
        if self.report_feat_imp is not None:
            self.report_feat_imp.plot(x="feature", y="feat_importance", kind="barh", title="Features Importance")
        else:
            raise ValueError("Missing report, run '.get_feature_importance()' first")



import pandas as pd
import numpy as np
from sklearn.metrics import recall_score


def lift_score(y_true, y_prob, method: str = "percentile", percentile: float = 0.2, threshold: float = 0.5):
    """Calculates the lift score which compares the model true positive rate versus a random choice.

    Args:
        y_true (np.array, pd.Series): real value of target
        y_prob (np.array, pd.Series): predicted probability of the target
        method (str, optional): 'percentile' or 'threshold'. If percentile it uses the defined percentiles to set the cutoff.
            If threshold it uses the defined threshold as cutoff. Default percentile
        percentile (float, optional): Percentile to use to get the cutoff. Defaults to 0.2.
        threshold (float, optional): Threshold to use as cutoff. Defaults to 0.5.

    Returns:
        float: lift score
    """
    checker_te = pd.DataFrame({"events": y_true, "prob": y_prob})
    if method == "percentile":
        cutoff = checker_te.prob.quantile(q=1 - percentile)
    else:
        cutoff = threshold
    if checker_te["events"].sum() == 0:
        raise ValueError("y_true always 0")

    tpr_db = checker_te.loc[checker_te.prob > cutoff]
    tpr_model = tpr_db["events"].sum() / tpr_db.shape[0]
    tpr_random = checker_te["events"].sum() / checker_te.shape[0]

    return tpr_model / tpr_random


def gain_score(y_true, y_prob, method: str = "percentile", percentile: float = 0.2, threshold: float = 0.5):
    """Calculates the gain score which identifies the recall at a cutoff.

    Args:
        y_true (np.array, pd.Series): real value of target
        y_prob (np.array, pd.Series): predicted probability of the target
        method (str, optional): 'percentile' or 'threshold'. If percentile it uses the defined percentiles to set the cutoff.
            If threshold it uses the defined threshold as cutoff. Default percentile
        percentile (float, optional): Percentile to use to get the cutoff. Defaults to 0.2.
        threshold (float, optional): Threshold to use as cutoff. Defaults to 0.5.

    Returns:
        float: gain score
    """
    checker_te = pd.DataFrame({"events": y_true, "prob": y_prob})
    if method == "percentile":
        cutoff = checker_te.prob.quantile(q=1 - percentile)
    else:
        cutoff = threshold
    if checker_te["events"].sum() == 0:
        raise ValueError("y_true always 0")
    return recall_score(y_true, np.where(checker_te.prob > cutoff, 1, 0))



from tqdm import tqdm
import sys

from model_monitoring.utils import (
    get_categorical_features,
    get_numerical_features,
    check_features_sets,
    convert_Int_dataframe,
)
from model_monitoring.reference_metadata.reference_metadata import retrieve_bins_dict, map_bins_dict


class ReferenceMetaData:
    """Reference Meta-Data Class."""

    def __init__(self, meta_ref_dict=None):
        """Reference Meta-Data Class.

        Args:
            meta_dict (dict, optional): reference metadata dictionary. Defaults to None.
        """
        self.meta_ref_dict = meta_ref_dict

    def get_meta_reference(self, data_reference, feat_to_check=None, nbins=1000, bin_min_pct=0.04, missing_values=True):
        """Retrieve the reference metadata dictionary.

        Args:
            data_reference (pd.DataFrame): original dataset for generating the reference metadata dictionary. Defaults to None.
            feat_to_check (list, optional): list of features to be checked. Deafualts to None.
            nbins (int, optional): number of bins into which the features will be bucketed (maximum). Defaults to 1000.
            bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.
            missing_values (bool, optional): boolean to add to the metadata dictionary information on missing values. Defaults to True.

        Returns:
            dict: reference metadata dictionary
        """
        if feat_to_check is None:
            feat_to_check = data_reference.columns

        self.feat_to_check = feat_to_check
        self.data_reference = convert_Int_dataframe(data_reference)

        self.nbins = nbins
        self.bin_min_pct = bin_min_pct
        self.missing_values = missing_values

        # Generation reference metadata dictionary
        meta_ref_dict = dict()
        numerical_feat = get_numerical_features(self.data_reference[self.feat_to_check])
        categorical_feat = get_categorical_features(self.data_reference[self.feat_to_check])

        features_pb = tqdm(self.feat_to_check, file=sys.stdout, desc="Performing bin mapping", ncols=100, leave=True)
        for ix, col in enumerate(features_pb):
            col_dict = dict()
            if col in numerical_feat:
                col_dict["type"] = "numerical"
                col_dict["min_val"] = self.data_reference[col].min()
                col_dict["max_val"] = self.data_reference[col].max()
                col_dict["not_missing_values"] = len(self.data_reference[col]) - self.data_reference[col].isnull().sum()
            if col in categorical_feat:
                col_dict["type"] = "categorical"
                col_dict["not_missing_values"] = len(self.data_reference[col]) - self.data_reference[col].isnull().sum()
            bins_dict = retrieve_bins_dict(self.data_reference, col, nbins=self.nbins, bin_min_pct=self.bin_min_pct)[
                col
            ]
            if bins_dict != dict():
                col_dict.update(bins_dict)
            if missing_values:
                col_dict["missing_values"] = self.data_reference[col].isnull().mean()
            meta_ref_dict[col] = col_dict

            if ix == len(self.feat_to_check) - 1:
                features_pb.set_description("Completed bin mapping", refresh=True)

        self.meta_ref_dict = meta_ref_dict

        return self.meta_ref_dict

    def get_meta_new(self, new_data, meta_dict=None):
        """Retrieve the metadata dictionary for new data using the reference metadata dictionary.

        Args:
            new_data (pd.DataFrame): new dataset.
            meta_dict (dict, optional): reference metadata dictionary. Defaults to None.

        Returns:
            dict: metadata dictionary for new data
        """
        if (meta_dict is None) and (self.meta_ref_dict is None):
            raise ValueError("no reference metadata dictionary provided")
        elif meta_dict is None:
            meta_dict = self.meta_ref_dict
        # Check if the features are the same
        check_features_sets(features_1=list(meta_dict.keys()), features_2=list(new_data.columns))

        list_com_metrics = list(set(meta_dict.keys()).intersection(set(new_data.columns)))
        self.new_data = convert_Int_dataframe(new_data[list_com_metrics])
        self.meta_dict = {k: meta_dict[k] for k in list_com_metrics}

        # Generation metadata dictionary for new data
        meta_new_dict = dict()
        for col in list_com_metrics:
            col_dict = dict()
            if self.meta_dict[col]["type"] not in ["categorical", "numerical"]:
                raise ValueError(
                    f"{self.meta_dict[col]['type']} is not a valid type for {col} feature. Choose between ['categorical','numerical']."
                )
            else:
                col_dict["type"] = self.meta_dict[col]["type"]
            if col_dict["type"] == "numerical":
                col_dict["min_val"] = self.new_data[col].min()
                col_dict["max_val"] = self.new_data[col].max()
            bins_dict = map_bins_dict(self.new_data, self.meta_dict, col)[col]
            if bins_dict != dict():
                col_dict.update(bins_dict)
            if "missing_values" in self.meta_dict[col]:
                col_dict["missing_values"] = self.new_data[col].isnull().mean()
                col_dict["not_missing_values"] = len(self.new_data[col]) - self.new_data[col].isnull().sum()
            meta_new_dict[col] = col_dict

        self.meta_new_dict = meta_new_dict

        return self.meta_new_dict



import pandas as pd
import warnings
from typing import Dict

from model_monitoring.utils import check_metrics_sets
from model_monitoring.config import read_config

from model_monitoring.fairness_drift.fairness_drift import (
    check_fairness_groups,
)

params = read_config(config_dir="config", name_params="params.yml")
standard_threshold = params["fairness_treshold"]


class FairnessDrift:
    """Fairness Drift Class."""

    def __init__(self, fair_metrics_curr, config_threshold=None):
        """Fairness Drift Class.

        Args:
            fair_metrics_curr (dict): dictionary containing current fairness metrics perfomances
            config_threshold (dict, optional): dictionary containing threshold settings. Defaults to None.
        """
        if not isinstance(fair_metrics_curr, Dict):
            raise ValueError(
                "Fairness metrics in input has not a valid format. It should be a dictionary containing functions as keys and values as values."
            )
        if config_threshold is None:
            config_threshold = standard_threshold

        check_metrics = [
            i for i in fair_metrics_curr.keys() if (i not in config_threshold.keys()) and (i != "proportion_1")
        ]
        if len(check_metrics) > 0:
            warnings.warn(f"{check_metrics} do not have threshold settings in config_threshold")

        list_com_metrics = list(set(fair_metrics_curr.keys()).intersection(set(config_threshold.keys())))
        fair_metrics_curr_com = {x: fair_metrics_curr[x] for x in list_com_metrics}

        # initialize reduced report with variables and groups
        self.report_reduced = pd.DataFrame(
            [
                {"vars": key, "groups": group, "curr_perc_label": values[1]}
                if len(values) == 2
                else {"vars": key, "groups": group, "curr_perc_label": values[1], "proportion_1_curr": values[2]}
                for key, groups in fair_metrics_curr_com[list(fair_metrics_curr_com.keys())[0]].items()
                for group, values in groups.items()
            ]
        )
        for x in fair_metrics_curr_com.keys():
            self.report_reduced[x + "_curr_perf"] = [
                values[0] for key, groups in fair_metrics_curr_com[x].items() for group, values in groups.items()
            ]
        # for output report columns ordering according to current percentage label
        self.report_reduced = self.report_reduced.sort_values(by=["vars", "curr_perc_label"], ascending=[True, False])

        self.fair_metrics_curr = fair_metrics_curr
        self.config_threshold = config_threshold
        self.perf_metrics_stor = None
        self.relative_reduced = False

    def get_absolute_reduced(self):
        """Load on the reduced report the absolute alert on current fairness metrics performances."""
        # Generation Alert
        for x in self.report_reduced.groups.values:
            warning_red = ""
            warning_yellow = ""
            for a in [y[:-10] for y in self.report_reduced.columns if "curr_perf" in y]:
                absolute_red = self.config_threshold[a]["absolute"]["red"]
                absolute_yellow = self.config_threshold[a]["absolute"]["yellow"]

                if self.config_threshold[a]["logic"] == "decrease":
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    if absolute_red != "None":
                        if curr_perf < absolute_red:
                            if warning_red == "":
                                warning_red += f"Red Alert for {a}"
                            else:
                                warning_red += f", {a}"
                        else:
                            if absolute_yellow != "None":
                                if (curr_perf > absolute_red) and (curr_perf < absolute_yellow):
                                    if warning_yellow == "":
                                        warning_yellow += f"Yellow Alert for {a}"
                                    else:
                                        warning_yellow += f", {a}"
                    else:
                        if absolute_yellow != "None":
                            if curr_perf < absolute_yellow:
                                if warning_yellow == "":
                                    warning_yellow += f"Yellow Alert for {a}"
                                else:
                                    warning_yellow += f", {a}"

                elif self.config_threshold[a]["logic"] == "increase":
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    if absolute_red != "None":
                        if curr_perf > absolute_red:
                            if warning_red == "":
                                warning_red += f"Red Alert for {a}"
                            else:
                                warning_red += f", {a}"
                        else:
                            if absolute_yellow != "None":
                                if (curr_perf < absolute_red) and (curr_perf > absolute_yellow):
                                    if warning_yellow == "":
                                        warning_yellow += f"Yellow Alert for {a}"
                                    else:
                                        warning_yellow += f", {a}"
                    else:
                        if absolute_yellow != "None":
                            if curr_perf > absolute_yellow:
                                if warning_yellow == "":
                                    warning_yellow += f"Yellow Alert for {a}"
                                else:
                                    warning_yellow += f", {a}"

                elif self.config_threshold[a]["logic"] == "axial":
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    if absolute_red != "None":
                        if (curr_perf > max(absolute_red)) or (curr_perf < min(absolute_red)):
                            if warning_red == "":
                                warning_red += f"Red Alert for {a}"
                            else:
                                warning_red += f", {a}"
                        else:
                            if absolute_yellow != "None":
                                if ((curr_perf < max(absolute_red)) and (curr_perf > min(absolute_red))) and (
                                    (curr_perf > max(absolute_yellow)) or (curr_perf < min(absolute_yellow))
                                ):
                                    if warning_yellow == "":
                                        warning_yellow += f"Yellow Alert for {a}"
                                    else:
                                        warning_yellow += f", {a}"
                    else:
                        if absolute_yellow != "None":
                            if (curr_perf > max(absolute_yellow)) or (curr_perf < min(absolute_yellow)):
                                if warning_yellow == "":
                                    warning_yellow += f"Yellow Alert for {a}"
                                else:
                                    warning_yellow += f", {a}"
                else:
                    raise ValueError(
                        f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                    )
            if (warning_red != "") and (warning_yellow != ""):
                warning = warning_red + ", " + warning_yellow
            else:
                warning = warning_red + warning_yellow
            self.report_reduced.loc[self.report_reduced.groups == x, "absolute_warning"] = warning

        # Locate Absolute_warning before Stor_Perc_label
        if self.relative_reduced:
            absolute_warning = self.report_reduced.pop("absolute_warning")

            self.report_reduced.insert(
                self.report_reduced.columns.get_loc("stor_perc_label"), "absolute_warning", absolute_warning
            )

    def get_relative_reduced(self, fair_metrics_stor):
        """Load on the reduced report the historichal fairness metrics performances, drift compared to the current fairness performances and relative alert on drift.

        Args:
            fair_metrics_stor (dict): dictionary containing historichal fairness metrics perfomances.
        """
        # re-initialize report
        if self.relative_reduced:
            list_drop = [
                x
                for x in self.report_reduced.columns
                if (x in ["stor_perc_label", "drift_perc", "relative_warning", "proportion_1_stor"])
                or ("stor_perf" in x)
            ]
            self.report_reduced = self.report_reduced.drop(columns=list_drop)

        # Check if the metrics are the same
        check_metrics_sets(metrics_1=fair_metrics_stor, metrics_2=self.fair_metrics_curr)

        list_com_metrics = list(set(fair_metrics_stor.keys()).intersection(set(self.fair_metrics_curr.keys())))
        self.fair_metrics_stor = {x: fair_metrics_stor[x] for x in list_com_metrics}
        stor_report_reduced = pd.DataFrame(
            [
                {"vars": key, "groups": group, "stor_perc_label": values[1]}
                if len(values) == 2
                else {"vars": key, "groups": group, "stor_perc_label": values[1], "proportion_1_stor": values[2]}
                for key, groups in self.fair_metrics_stor[list(self.fair_metrics_stor.keys())[0]].items()
                for group, values in groups.items()
            ]
        )
        for x in self.fair_metrics_stor.keys():
            stor_report_reduced[x + "_stor_perf"] = [
                values[0] for key, groups in self.fair_metrics_stor[x].items() for group, values in groups.items()
            ]

        # Check if the fairness group are the same
        list_no_join = check_fairness_groups(self.report_reduced, stor_report_reduced, multiindex=False)

        # Add historical fairness performances to the reduced report and limit to common fairness groups
        self.report_reduced = self.report_reduced.merge(stor_report_reduced, how="inner", on=["vars", "groups"])

        # Generation Drift
        for a in [y[:-10] for y in self.report_reduced.columns if "stor_perf" in y]:
            if self.config_threshold[a]["logic"] in ["decrease", "increase"]:
                for x in self.report_reduced.groups.values:
                    stor_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_stor_perf"].values[0]
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    if stor_perf > 0:
                        self.report_reduced.loc[self.report_reduced.groups == x, "drift_perc_" + a] = (
                            (curr_perf - stor_perf) / stor_perf * 100
                        )
                    else:
                        self.report_reduced.loc[self.report_reduced.groups == x, "drift_perc_" + a] = (
                            (stor_perf - curr_perf) / stor_perf * 100
                        )
                drift = self.report_reduced.pop("drift_perc_" + a)
                self.report_reduced.insert(
                    self.report_reduced.columns.get_loc(a + "_stor_perf") + 1, "drift_perc_" + a, drift
                )

            elif self.config_threshold[a]["logic"] == "axial":
                axial_point = self.config_threshold[a]["axial_point"]
                for x in self.report_reduced.groups.values:
                    stor_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_stor_perf"].values[0]
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    self.report_reduced.loc[self.report_reduced.groups == x, "drift_perc_" + a] = (
                        (abs(curr_perf - axial_point) - abs(stor_perf - axial_point))
                        / abs(stor_perf - axial_point)
                        * 100
                    )
                drift = self.report_reduced.pop("drift_perc_" + a)
                self.report_reduced.insert(
                    self.report_reduced.columns.get_loc(a + "_stor_perf") + 1, "drift_perc_" + a, drift
                )

            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                )

        # Generation Alert
        for x in self.report_reduced.groups.values:
            warning_red = ""
            warning_yellow = ""
            for a in [y[:-10] for y in self.report_reduced.columns if "curr_perf" in y]:
                relative_red = self.config_threshold[a]["relative"]["red"]
                relative_yellow = self.config_threshold[a]["relative"]["yellow"]
                absolute_tol = self.config_threshold[a]["relative"]["absolute_tol"]

                if self.config_threshold[a]["logic"] == "decrease":
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    stor_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_stor_perf"].values[0]
                    if relative_red != "None":
                        # check absolute tollerance for relative alert
                        if abs(curr_perf - stor_perf) >= absolute_tol:
                            drift_perf = self.report_reduced.loc[
                                self.report_reduced.groups == x, "drift_perc_" + a
                            ].values[0]
                            if drift_perf < relative_red * 100:
                                if warning_red == "":
                                    warning_red += f"Red Alert for {a}"
                                else:
                                    warning_red += f", {a}"
                            else:
                                if relative_yellow != "None":
                                    if (drift_perf > relative_red * 100) and (drift_perf < relative_yellow * 100):
                                        if warning_yellow == "":
                                            warning_yellow += f"Yellow Alert for {a}"
                                        else:
                                            warning_yellow += f", {a}"
                    else:
                        if relative_yellow != "None":
                            # check absolute tollerance for relative alert
                            if abs(curr_perf - stor_perf) >= absolute_tol:
                                drift_perf = self.report_reduced.loc[
                                    self.report_reduced.groups == x, "drift_perc_" + a
                                ].values[0]
                                if drift_perf < relative_yellow * 100:
                                    if warning_yellow == "":
                                        warning_yellow += f"Yellow Alert for {a}"
                                    else:
                                        warning_yellow += f", {a}"

                elif self.config_threshold[a]["logic"] in ["increase", "axial"]:
                    curr_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_curr_perf"].values[0]
                    stor_perf = self.report_reduced.loc[self.report_reduced.groups == x, a + "_stor_perf"].values[0]
                    if relative_red != "None":
                        # check absolute tollerance for relative alert
                        if abs(curr_perf - stor_perf) >= absolute_tol:
                            drift_perf = self.report_reduced.loc[
                                self.report_reduced.groups == x, "drift_perc_" + a
                            ].values[0]
                            if drift_perf > relative_red * 100:
                                if warning_red == "":
                                    warning_red += f"Red Alert for {a}"
                                else:
                                    warning_red += f", {a}"
                            else:
                                if relative_yellow != "None":
                                    if (drift_perf < relative_red * 100) and (drift_perf > relative_yellow * 100):
                                        if warning_yellow == "":
                                            warning_yellow += f"Yellow Alert for {a}"
                                        else:
                                            warning_yellow += f", {a}"
                    else:
                        if relative_yellow != "None":
                            # check absolute tollerance for relative alert
                            if abs(curr_perf - stor_perf) >= absolute_tol:
                                drift_perf = self.report_reduced.loc[
                                    self.report_reduced.groups == x, "drift_perc_" + a
                                ].values[0]
                                if drift_perf > relative_yellow * 100:
                                    if warning_yellow == "":
                                        warning_yellow += f"Yellow Alert for {a}"
                                    else:
                                        warning_yellow += f", {a}"

                else:
                    raise ValueError(
                        f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                    )
            if (warning_red != "") and (warning_yellow != ""):
                warning = warning_red + ", " + warning_yellow
            else:
                warning = warning_red + warning_yellow
            self.report_reduced.loc[self.report_reduced.groups == x, "relative_warning"] = warning

        self.relative_reduced = True

    def get_report_reduced(self):
        """Return the reduced report.

        Returns:
            pd.DataFrame: reduced report of the class
        """
        return self.report_reduced



import pandas as pd


def bg_red_yellow(df):
    """Color report of Class Performance Drift based on Alert.

    Args:
        df (pd.DataFrame): report in input

    Returns:
        pd.DataFrame: color-mapping report
    """
    ret = pd.DataFrame("", index=df.index, columns=df.columns)
    if "absolute_warning" in df.columns:
        ret.loc[df.absolute_warning == "Red Alert", ["curr_perf", "absolute_warning"]] = "background-color: red"
        ret.loc[df.absolute_warning == "Yellow Alert", ["curr_perf", "absolute_warning"]] = "background-color: yellow"

    ret.loc[df.relative_warning == "Red Alert", ["drift_perc", "relative_warning"]] = "background-color: red"
    ret.loc[df.relative_warning == "Yellow Alert", ["drift_perc", "relative_warning"]] = "background-color: yellow"
    return ret



from setuptools import setup, find_packages

setup(
    name="model_monitoring",
    version="1.1.20",
    description="Package for Model Monitoring",
    author="DAT Team",
    url="https://dev.azure.com/credem-data/DAT/_git/model_monitoring",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=find_packages(where="src"),
    package_data={
        "model_monitoring": ["config/*.yml"],
    },
    include_package_data=True,
    package_dir={"": "src"},
    # Usato solo con 3.9 e 3.10
    python_requires=">=3.9, <4",
    install_requires=[
        "typed-ast==1.5.4",
        "numpy==1.22.4",
        "pandas==1.5.3",
        "scikit-learn==1.2.1",
        "lightgbm==3.3.2",
        "scipy==1.10.1",
        "PyYAML==6.0",
        "shap==0.41.0",
        "ipython",
    ],
)



import pandas as pd
import numpy as np
import warnings


def check_size(data_1, data_2):
    """Check if two np.arrays/pd.DataFrames/pd.Series have the same length.

    Args:
        data_1 (np.array/pd.Series/pd.DataFrame): data 1
        data_2 (np.array/pd.Series/pd.DataFrame): data 2

    Returns:
        ValueError: ValueError if data_1 and data_2 do not have the same length
    """
    if len(data_1) != len(data_2):
        raise ValueError("input dataset do not have the same length ({0},{1})".format(len(data_1), len(data_2)))


def check_metrics_sets(metrics_1: dict, metrics_2: dict):
    """Check that metrics in set 1 are the same of those in set 2. If not raise a warning.

    Args:
        metrics_1 (dict): dictionary of metrics 1
        metrics_2 (dict): dictionary of metrics 2
    """
    metrics = set(metrics_1.keys())
    cnf_metrics = set(metrics_2.keys())

    if metrics ^ cnf_metrics != set():
        warnings.warn(f"unmatched metrics {metrics^cnf_metrics}")


def check_features_sets(features_1: list, features_2: list):
    """Check that features in set 1 are the same of those in set 2. If not raise a warning.

    Args:
        features_1 (list): list of features 1
        features_2 (list): list of features 2
    """
    features_base = set(features_1)
    features_compare = set(features_2)

    if features_base ^ features_compare != set():
        warnings.warn(f"unmatched features {features_base^features_compare}. Analysis is computed on common features.")


def get_categorical_features(df, feature_filter=None):
    """It gets a list of categorical features' names from a DataFrame excluding unwanted features' names.

    Args:

    df (pd.DataFrame): input dataframe
    feature_filter (list, optional): list of columns that will be filtered when extracting
        categorical features.

    Returns:
        list: the list of categorical feature names
    """
    if feature_filter is None:
        feature_filter = []

    categorical_feat = [
        var for var in df.columns if (df[var].dtype == "O" or df[var].dtype == "category") and var not in feature_filter
    ]

    return categorical_feat


def get_numerical_features(df, feature_filter=None):
    """It gets a list of numerical features' names from a DataFrame excluding unwanted features' names.

    Args:
        df (pd.DataFrame): input dataframe
        feature_filter (list, optional): list of columns that will be filtered when extracting
            numerical features.

    Returns:
        list: the list of numerical feature names
    """
    if feature_filter is None:
        feature_filter = []

    numerical_feat = [
        var
        for var in df.columns
        if (df[var].dtype != "O" and df[var].dtype != "category") and (var not in feature_filter)
    ]
    return numerical_feat


def merge_categorical_bins(df, feat, mapper, bin_min_pct=0.04):
    """Merge buckets that have low frequency.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        mapper (dict): initial mapper to be used to bucket the feature feat
        bin_min_pct (float, optional): minimum percentage of observations per bucket. Defaults to 0.04.

    Returns:
        dict: mapper to be used to bucket the feature feat
    """
    db = df[[feat]].dropna().assign(bin=df[feat].dropna().map(mapper))

    freqs = db["bin"].value_counts(ascending=True, normalize=True)

    if freqs.iloc[0] < bin_min_pct:
        update = {x: freqs.index[1] for x, y in mapper.items() if y == freqs.index[0]}
        mapper.update(update)
        mapper = merge_categorical_bins(df, feat, mapper, bin_min_pct)

    return mapper


def retrieve_bin_numerical(df, feat, max_n_bins=1000):
    """Retrieve the bucket for a numerical feature.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        max_n_bins (int, optional): number of bins into which the features will be bucketed (maximum) to compute psi
    Returns:
        list: the list of cuts to be used in pd.cut
    """
    if df[feat].dropna().nunique() > max_n_bins:
        db = (
            df[[feat]].dropna().assign(bucket=pd.qcut(df[feat].dropna(), q=max_n_bins, labels=False, duplicates="drop"))
        )
        cuts = list(db.groupby("bucket")[feat].max())
    else:
        cuts = list(df[feat].dropna().unique())

    cuts.sort()
    cuts = [-np.Inf] + cuts
    cuts[-1] = np.Inf

    return cuts


def merge_numerical_bins(df, feat, cuts, bin_min_pct=0.04):
    """Merge buckets that have low frequency.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        cuts (list): initial cuts to be used to bucket the feature feat
        bin_min_pct (float, optional): minimum percentage of observations per bucket

    Returns:
        dict: mapper to be used to bucket the feature feat
    """
    db = df[[feat]].dropna().assign(bin=pd.cut(df[feat].dropna(), bins=cuts, labels=False, right=True))
    freqs = db["bin"].value_counts(ascending=True, normalize=True).sort_index()

    n = freqs.index[-1]
    j = 0
    for i in range(n + 1):
        if freqs.iloc[i - j] < bin_min_pct:
            j += 1
            if (i + 1 - j) == freqs.index[-1]:
                remove_cut = cuts[-2]
            else:
                remove_cut = cuts[i + 2 - j]
            cuts.remove(remove_cut)
            db = df[[feat]].dropna().assign(bin=pd.cut(df[feat].dropna(), bins=cuts, labels=False, right=True))
            freqs = db["bin"].value_counts(ascending=True, normalize=True).sort_index()

    return cuts


def convert_Int_dataframe(db):
    """Converts feature "Int" format into "float" format of the dataset in input.

    Args:
        db (pd.DataFrame): dataset for integer converting mapping

    Returns:
        pd.DataFrame: dataframe with converted int types features
    """
    return db.astype(
        {
            **{col: "float64" for col in db.columns if db[col].dtype == "Int64"},
            **{col: "float32" for col in db.columns if db[col].dtype == "Int32"},
        }
    )


def convert_Int_series(db):
    """Converts feature "Int" format into "float" format of the dataset in input.

    Args:
        db (pd.Series): pd.Series for integer converting mapping

    Returns:
        pd.Series: series with converted int types features
    """
    return db.astype("float64") if db.dtype == "Int64" else db.astype("float32") if db.dtype == "Int32" else db


def absmax(a, axis=0):
    """Retrieve the maximum values in absolute value along an axis from a numpy ndarray.

    Args:
        a (np.ndarray): numpy ndarray from which we get the maximum in absolute value
        axis (int): determines the axis along which we calculate the maximum in absolute value

    Returns:
        np.array: np.array with the maximum values in asbolute value along an axis
    """
    s = np.array(a.shape)
    s[axis] = -1
    return np.take_along_axis(a, np.abs(a).argmax(axis).reshape(s), axis=axis)[0]
